from __future__ import annotations

from collections import defaultdict

from capstone6pwndbg import *  # noqa: F403

import pwndbg
import pwndbg.aglib
import pwndbg.aglib.disasm.disassembly
import pwndbg.aglib.memory
import pwndbg.aglib.symbol
import pwndbg.aglib.vmmap
import pwndbg.color
import pwndbg.color.context as ctx_color
import pwndbg.color.disasm
import pwndbg.color.theme
import pwndbg.commands.comments
import pwndbg.lib.config
from pwndbg.aglib.disasm.instruction import PwndbgInstruction
from pwndbg.aglib.disasm.instruction import SplitType
from pwndbg.color import ColorConfig
from pwndbg.color import ColorParamSpec
from pwndbg.color import blue
from pwndbg.color import green
from pwndbg.color import light_gray
from pwndbg.color import light_green
from pwndbg.color import light_purple
from pwndbg.color import light_red
from pwndbg.color import message
from pwndbg.color import purple
from pwndbg.color import red
from pwndbg.color import rjust_colored
from pwndbg.color import white
from pwndbg.color import yellow


def ljust_padding(lst: list[str]) -> tuple[list[str], int]:
    """
    Returns (list of padded strings, max length of string)
    """
    longest_len = max(map(len, lst)) if lst else 0
    return [s.ljust(longest_len) for s in lst], longest_len


c = ColorConfig(
    "nearpc",
    [
        ColorParamSpec("symbol", "normal", "color for nearpc command (symbol)"),
        ColorParamSpec("address", "normal", "color for nearpc command (address)"),
        ColorParamSpec("prefix", "none", "color for nearpc command (prefix marker)"),
        ColorParamSpec("breakpoint", "red", "color for nearpc command (breakpoint marker)"),
        ColorParamSpec("syscall-name", "red", "color for nearpc command (resolved syscall name)"),
        ColorParamSpec("argument", "bold", "color for nearpc command (target argument)"),
        ColorParamSpec(
            "integration-comments", "bold", "color for nearpc command (integration comments)"
        ),
        ColorParamSpec("branch-marker", "normal", "color for nearpc command (branch marker line)"),
    ],
)

# `pwndbg.arguments` imports `c` from this module.
import pwndbg.arguments

nearpc_branch_marker = pwndbg.color.theme.add_param(
    "nearpc-branch-marker", "    ↓", "branch marker line for nearpc command"
)
nearpc_branch_marker_contiguous = pwndbg.color.theme.add_param(
    "nearpc-branch-marker-contiguous",
    " ",
    "contiguous branch marker line for nearpc command",
)
pwndbg.color.theme.add_param("highlight-pc", True, "whether to highlight the current instruction")
pwndbg.color.theme.add_param("highlight-breakpoints", True, "whether to highlight breakpoints")
pwndbg.color.theme.add_param("nearpc-prefix", "►", "prefix marker for nearpc command")
pwndbg.color.theme.add_param(
    "nearpc-breakpoint-prefix", "b+", "breakpoint marker for nearpc command"
)
pwndbg.config.add_param("left-pad-disasm", True, "whether to left-pad disassembly")
show_args = pwndbg.config.add_param(
    "nearpc-show-args", True, "whether to show call arguments below instruction"
)
# show_comments = pwndbg.config.add_param(
#     "nearpc-integration-comments",
#     True,
#     "whether to show comments from integration provider",
# )
show_opcode_bytes = pwndbg.config.add_param(
    "nearpc-num-opcode-bytes",
    0,
    "number of opcode bytes to print for each instruction",
    param_class=pwndbg.lib.config.PARAM_ZUINTEGER,
)
opcode_separator_bytes = pwndbg.config.add_param(
    "nearpc-opcode-separator-bytes",
    1,
    "number of spaces between opcode bytes",
    param_class=pwndbg.lib.config.PARAM_ZUINTEGER,
)


class JumpRange:
    start: int
    end: int
    forward: bool

    min: int
    max: int

    def __init__(self, s: int, e: int):
        self.start = s
        self.end = e

        self.forward = self.start < self.end

        self.min = min(self.start, self.end)
        self.max = max(self.start, self.end)

    def contains(self, address: int) -> bool:
        return self.min <= address <= self.max

    def overlaps(self, other: JumpRange) -> bool:
        return max(self.min, other.min) <= min(self.max, other.max)

    def __repr__(self) -> str:
        return f"JumpRange({hex(self.start)}, {hex(self.end)})"


COLUMNS_ALLOCATED_FOR_BRANCH_VISUALIZATION = 20

# Symbols used in branch visualization
TOP_LEFT_CORNER = "┌"
BOT_LEFT_CORNER = "└"
HORZ_SYMBOL = "─"
VERT_SYMBOL = "│"
START_SYMBOL = "<"
END_SYMBOL = ">"
DOTTED_VERTICAL = "╎"
UP_SYMBOL = "▲"

offset_to_color_map = {
    0: white,
    1: red,
    2: green,
    3: purple,
    4: blue,
    5: white,
    6: yellow,
    7: light_red,
    8: light_purple,
    9: light_gray,
    10: light_green,
}


# Allows to the branch visualization work across repeated uses of nearpc
# Maps the jump range to the id it was given.
last_run_ids: dict[JumpRange, int] = {}


def colorize_branch_vis_line(offset: int, string: str):
    return offset_to_color_map.get(offset, lambda x: str(x))(string)


def preprocess_branch_visualization(
    instructions: list[PwndbgInstruction], repeat: bool
) -> tuple[dict[int, list[JumpRange]], dict[JumpRange, int], int]:
    """
    Returns (pair_map dictionary,pair_id dictionary, maximum_pair_id)
    """
    global last_run_ids

    jumps: list[JumpRange] = []

    # Map of every address to every jump range it belongs in
    pair_map: dict[int, list[JumpRange]] = defaultdict(list)

    # For a given jump range, it has a unique ID compared to every other jump range it overlaps with
    # This allows us to display each jump range at a different visual offset
    pair_id: dict[JumpRange, int] = defaultdict(lambda: -1)

    # -2 because at least two columns are needed: one for the "<" or ">" branch ends, and one for pair id 0
    maximum_pair_id = COLUMNS_ALLOCATED_FOR_BRANCH_VISUALIZATION - 2

    # Find all instructions eligible for branch visualization
    for instruction in instructions:
        if instruction.jump_like and instruction.has_jump_target and not instruction.call_like:
            jumps.append(JumpRange(instruction.address, instruction.target))

    # Of the jumpranges we processed last time, which ones do we keep? Relevant for repeat nearpc
    continued_ranges: set[JumpRange] = set()

    # Population structure mapping every address to each jump range it belongs to
    for instruction in instructions:
        for pair in jumps:
            if pair.contains(instruction.address):
                pair_map[instruction.address].append(pair)

        if repeat:
            for pair, y in last_run_ids.items():
                if pair.contains(instruction.address):
                    pair_map[instruction.address].append(pair)
                    continued_ranges.add(pair)

    if repeat:
        for pair in continued_ranges:
            pair_id[pair] = last_run_ids[pair]
            jumps.append(pair)

    # Preprocess each pair to assign a unique ID to all overlapping ranges.
    for pair1 in jumps:
        # If this was from a repeat range, ignore it
        if pair_id[pair1] >= 0:
            continue

        cur_offset = 0
        for pair2 in jumps:
            if pair1 == pair2:
                continue

            if pair1.overlaps(pair2):
                # These two jump ranges overlap! Make sure pair1 has a larger offset!
                if pair_id[pair2] >= cur_offset:
                    cur_offset = pair_id[pair2] + 1

        # We only want a maximum number of columns
        pair_id[pair1] = min(cur_offset, maximum_pair_id)

    # Sort lists of jump ranges by ascending id
    for instruction in instructions:
        pairs = pair_map[instruction.address]
        pairs.sort(key=lambda x: pair_id[x])

    last_run_ids = pair_id

    return pair_map, pair_id, maximum_pair_id


def create_branch_visualization_strings(
    pair_map: dict[int, list[JumpRange]],
    pair_id: dict[JumpRange, int],
    maximum_pair_id: int,
    addr: int,
    is_first_address: bool,
) -> tuple[str, str]:
    """
    Returns tuple of (string, string for empty line)
    """

    if len(pair_id) == 0:
        return None, None

    # This string has ANSI colors in it
    branch_vis_string = ""

    # Length of the string ignoring ANSI color codes
    branch_vis_string_len = 0

    # This is the string placed in to the empty lines after branches
    empty_line_branch_vis_string = ""
    empty_line_branch_vis_string_len = 0

    # First, handle creating the horizontal lines (handling all the jumps that are start or end here)
    for pair in pair_map[addr]:
        # Due to preprocessing, we are iterating jump ranges at this address in order of smallest to largest id
        pair_offset = pair_id[pair]

        # The number of new columns to create in the string in this pass
        expand_amount = min(pair_offset, pair_offset - branch_vis_string_len + 1)

        # If a forward jump
        if pair.forward:
            if pair.start == addr:
                if branch_vis_string:
                    branch_vis_string = (
                        colorize_branch_vis_line(
                            pair_offset, TOP_LEFT_CORNER + (expand_amount) * HORZ_SYMBOL
                        )
                        + branch_vis_string
                    )
                    branch_vis_string_len += 1 + expand_amount
                else:
                    branch_vis_string = colorize_branch_vis_line(
                        pair_offset,
                        TOP_LEFT_CORNER + (expand_amount) * HORZ_SYMBOL + START_SYMBOL,
                    )
                    branch_vis_string_len += 2 + expand_amount
            elif pair.end == addr:
                if branch_vis_string:
                    branch_vis_string = (
                        colorize_branch_vis_line(
                            pair_offset, BOT_LEFT_CORNER + (expand_amount) * HORZ_SYMBOL
                        )
                        + branch_vis_string
                    )
                    branch_vis_string_len += 1 + expand_amount
                else:
                    branch_vis_string = colorize_branch_vis_line(
                        pair_offset,
                        BOT_LEFT_CORNER + (expand_amount) * HORZ_SYMBOL + END_SYMBOL,
                    )
                    branch_vis_string_len += 2 + expand_amount
        else:
            # Backwards jump
            if pair.start == addr:
                if branch_vis_string:
                    branch_vis_string = (
                        colorize_branch_vis_line(
                            pair_offset, BOT_LEFT_CORNER + (expand_amount) * HORZ_SYMBOL
                        )
                        + branch_vis_string
                    )
                    branch_vis_string_len += 1 + expand_amount
                else:
                    branch_vis_string = colorize_branch_vis_line(
                        pair_offset,
                        BOT_LEFT_CORNER + (expand_amount) * HORZ_SYMBOL + START_SYMBOL,
                    )
                    branch_vis_string_len += 2 + expand_amount
            elif pair.end == addr:
                if branch_vis_string:
                    branch_vis_string = (
                        colorize_branch_vis_line(
                            pair_offset, TOP_LEFT_CORNER + (expand_amount) * HORZ_SYMBOL
                        )
                        + branch_vis_string
                    )
                    branch_vis_string_len += 1 + expand_amount
                else:
                    branch_vis_string = colorize_branch_vis_line(
                        pair_offset,
                        TOP_LEFT_CORNER + (expand_amount) * HORZ_SYMBOL + END_SYMBOL,
                    )
                    branch_vis_string_len += 2 + expand_amount
        if pair_offset == maximum_pair_id:
            # We don't have any more column space for more jump ranges
            break

    # Secondly, handle the vertical lines passing through this address

    # This loop has multiple ways exit due to reaching the full column space
    # It's easier to track this with a variable.
    last_iteration = False

    for pair in pair_map[addr]:
        if last_iteration:
            break

        pair_offset = pair_id[pair]

        vert_symbol = empty_line_vert_symbol = VERT_SYMBOL

        if not pair.forward:
            empty_line_vert_symbol = DOTTED_VERTICAL
            if not is_first_address:
                vert_symbol = DOTTED_VERTICAL
            else:
                # Handle edge case: during repeated nearpc, entering a region that has a jump that goes backwards
                vert_symbol = UP_SYMBOL

        # First, handle creating the vertical lines that pass through the empty row created after branches (if it exists)
        if pair.forward:
            # If this pair ended at this address, nothing goes into the empty line after it
            if pair.end == addr:
                continue
        else:
            # If a backwards jump started here, nothing goes in the empty line after it
            if pair.start == addr:
                continue

        if pair_offset == maximum_pair_id:
            last_iteration = True

        # A single column is always taken for the < or > characters
        target_column = pair_offset + 1

        num_empty_lines = min(target_column, target_column - empty_line_branch_vis_string_len)
        empty_line_branch_vis_string = (
            colorize_branch_vis_line(
                pair_offset,
                empty_line_vert_symbol + (" " * num_empty_lines),
            )
            + empty_line_branch_vis_string
        )
        empty_line_branch_vis_string_len += 1 + num_empty_lines

        # Now, create the string for the non-empty line
        if pair.start == addr or pair.end == addr:
            continue

        # Only add to the string if the space hasn't been taking by a horizontal line
        if branch_vis_string_len <= target_column:
            empty_lines = min(target_column, target_column - branch_vis_string_len)
            branch_vis_string = (
                colorize_branch_vis_line(
                    pair_offset,
                    vert_symbol + (" " * empty_lines),
                )
                + branch_vis_string
            )
            branch_vis_string_len += 1 + empty_lines

    branch_vis_string = rjust_colored(branch_vis_string, COLUMNS_ALLOCATED_FOR_BRANCH_VISUALIZATION)
    empty_line_branch_vis_string = rjust_colored(
        empty_line_branch_vis_string, COLUMNS_ALLOCATED_FOR_BRANCH_VISUALIZATION
    )

    return branch_vis_string, empty_line_branch_vis_string


def nearpc(
    pc: int | None = None,
    lines: int | None = None,
    back_lines: int = 0,
    total_lines: int | None = None,
    emulate=False,
    repeat=False,
    use_cache=False,
    linear=False,
    branch_visualization=False,
    end_address: int | None = None,
) -> list[str]:
    """
    Disassemble near a specified address.

    The `linear` argument specifies if we should disassemble linearly in memory, or take jumps into account
    """

    # Repeating nearpc (pressing enter) makes it show next addresses
    # (writing nearpc explicitly again will reset its state)
    if repeat:
        # TODO: It would be better to do this in the nearpc command itself, but
        # that would require a larger refactor
        pc = nearpc.next_pc

    result: list[str] = []

    if pc is not None:
        pc = pwndbg.dbg.selected_inferior().create_value(pc).cast(pwndbg.aglib.typeinfo.pvoid)

    if pc is None:
        pc = pwndbg.aglib.regs.pc

    pc = int(pc)

    # Check whether we can even read this address
    if not pwndbg.aglib.memory.peek(pc):
        result.append(message.error(f"Invalid address {pc:#x}"))

    if lines is None:
        lines = int(pwndbg.config.nearpc_lines)

    # # Load source data if it's available
    # pc_to_linenos = collections.defaultdict(lambda: [])
    # lineno_to_src = {}
    # frame = gdb.selected_frame()
    # if frame:
    #     sal = frame.find_sal()
    #     if sal:
    #         symtab = sal.symtab
    #         objfile = symtab.objfile
    #         sourcefilename = symtab.filename
    #         with open(sourcefilename, 'r') as sourcefile:
    #             lineno_to_src = {i:l for i,l in enumerate(sourcefile.readlines())}

    #         for line in symtab.linetable():
    #             pc_to_linenos[line.pc].append(line.line)

    instructions, index_of_pc = pwndbg.aglib.disasm.disassembly.near(
        pc,
        forward_count=lines,
        backward_count=back_lines,
        total_count=total_lines,
        emulate=emulate,
        show_prev_insns=not repeat,
        use_cache=use_cache,
        linear=linear,
        end_address=end_address,
    )

    # If doing branch visualization, preprocess some datastructures
    if branch_visualization:
        pair_map, pair_id, maximum_pair_id = preprocess_branch_visualization(instructions, repeat)

    if pwndbg.aglib.memory.peek(pc) and not instructions:
        result.append(message.error(f"Invalid instructions at {pc:#x}"))

    # In case $pc is in a new map we don't know about,
    # this will trigger an exploratory search.
    pwndbg.aglib.vmmap.find(pc)

    # Gather all addresses and symbols for each instruction
    # Ex: <main+43>
    symbols = [pwndbg.aglib.symbol.resolve_addr(i.address) for i in instructions]
    addresses: list[str] = [f"{i.address:#x}" for i in instructions]

    nearpc.next_pc = instructions[-1].address + instructions[-1].size if instructions else 0

    # Format the symbol name for each instruction
    symbols = [f"<{sym}> " if sym else "" for sym in symbols]

    # Pad out all of the symbols and addresses
    if pwndbg.config.left_pad_disasm:
        symbols, symbols_max_length = ljust_padding(symbols)
        addresses, addresses_max_length = ljust_padding(addresses)
    else:
        symbols_max_length = max(map(len, symbols)) if symbols else 0
        addresses_max_length = max(map(len, addresses)) if addresses else 0

    assembly_strings = pwndbg.color.disasm.instructions_and_padding(instructions)

    breakpoint_locations = pwndbg.dbg.breakpoint_locations()

    prefix_sign = pwndbg.config.nearpc_prefix
    current_insn_prefix = f" {prefix_sign}"
    current_insn_prefix = c.prefix(current_insn_prefix)
    default_prefix = " " * (len(prefix_sign) + 1)
    default_prefix = c.prefix(default_prefix)

    breakpoint_sign = pwndbg.config.nearpc_breakpoint_prefix
    breakpoint_prefix = breakpoint_sign.ljust(len(prefix_sign) + 1)
    breakpoint_prefix = c.breakpoint(breakpoint_prefix)

    # Print out each instruction
    for i, (address_str, symbol, instruction, asm) in enumerate(
        zip(addresses, symbols, instructions, assembly_strings)
    ):
        # Show prefix only on the specified address and don't show it while in repeat-mode
        # or when showing current instruction for the second time
        show_prefix = instruction.address == pc and not repeat and i == index_of_pc
        is_breakpoint = False
        if show_prefix:
            prefix = current_insn_prefix
        elif instruction.address in breakpoint_locations:
            # If the instruction is not the current instruction and a breakpoint,
            # show the breakpoint sign
            prefix = breakpoint_prefix
            is_breakpoint = True
        else:
            prefix = default_prefix

        # If this instruction is a breakpoint and not the current pc, highlight it.
        if is_breakpoint and pwndbg.config.highlight_breakpoints:
            address_str = c.breakpoint(address_str)
            symbol = c.breakpoint(symbol)
        # Colorize address and symbol if not highlighted
        # symbol is fetched from gdb and it can be e.g. '<main+8>'
        # In case there are duplicate instances of an instruction (tight loop),
        # ones that the instruction pointer is not at stick out a little, to indicate the repetition
        elif not pwndbg.config.highlight_pc or instruction.address != pc or repeat:
            address_str = c.address(address_str)
            symbol = c.symbol(symbol)
        elif pwndbg.config.highlight_pc and i == index_of_pc:
            # If this instruction is the one the PC is at.
            # In case of tight loops, with emulation we may display the same instruction multiple times.
            # Only highlight current instance, not past or future times.
            address_str = ctx_color.highlight(address_str)
            symbol = ctx_color.highlight(symbol)

        # If this instruction performs a memory access operation, we should tell
        # the user anything we can figure out about the memory it's trying to
        # access.
        # mem_access = ""
        if instruction.address == pc and False:
            accesses = []
            for operand in instruction.operands:
                if operand.type != CS_OP_MEM:
                    continue
                address = operand.mem.disp

                base = operand.mem.base
                if base > 0:
                    address += pwndbg.aglib.regs.read_reg(instruction.reg_name(base))

                vmmap = pwndbg.aglib.vmmap.get()
                page = next((page for page in vmmap if address in page), None)
                if page is None:
                    # This is definetly invalid. Don't even bother checking
                    # any other conditions.
                    accesses.append(f"[X] {address:#x}")
                    continue

                if operand.access == CS_AC_READ and not page.read:
                    # Tried to read from a page we can't read.
                    accesses.append(f"[X] {address:#x}")
                    continue
                if operand.access == CS_AC_WRITE and not page.write:
                    # Tried to write to a page we can't write.
                    accesses.append(f"[X] {address:#x}")
                    continue

                # At this point, we know the operation is legal, but we don't
                # know where it's going yet. It could be going to either memory
                # managed by libc or memory managed by the program itself.

                if not pwndbg.aglib.heap.current.is_initialized():
                    # The libc heap hasn't been initialized yet. There's not a
                    # lot that we can say beyond this point.
                    continue
                allocator = pwndbg.aglib.heap.current

                heap = pwndbg.aglib.heap.ptmalloc.Heap(address)
                chunk = None
                for ch in heap:
                    # Find the chunk in this heap the corresponds to the address
                    # we're trying to access.
                    offset = address - ch.address
                    if offset >= 0 and offset < ch.real_size:
                        chunk = ch
                        break
                if chunk is None:
                    # The memory for this chunk is not managed by libc. We can't
                    # reason about it.
                    accesses.append(f"[?] {address:#x}")
                    continue

                # Scavenge through all of the bins in the current allocator.
                # Bins track free chunks, so, whether or not we can find the
                # chunk we're trying to access in a bin will tells us whether
                # this access is a UAF.
                bins_list = [
                    allocator.fastbins(chunk.arena.address),
                    allocator.smallbins(chunk.arena.address),
                    allocator.largebins(chunk.arena.address),
                    allocator.unsortedbin(chunk.arena.address),
                ]
                if allocator.has_tcache():
                    bins_list.append(allocator.tcachebins(None))

                bins_list = [x for x in bins_list if x is not None]
                for bins in bins_list:
                    if bins.contains_chunk(chunk.real_size, chunk.address):
                        # This chunk is free. This is a UAF.
                        accesses.append(f"[UAF] {address:#x}")
                        continue
            # mem_access = " ".join(accesses)

        opcodes = ""
        if show_opcode_bytes > 0:
            opcodes = (opcode_separator_bytes * " ").join(
                f"{c:02x}" for c in instruction.bytes[: int(show_opcode_bytes)]
            )
            # Must add +3 at minimum, due to truncated instructions adding "..."
            align = show_opcode_bytes * 2 + 3
            if opcode_separator_bytes > 0:
                # add the length of the maximum number of separators to the alignment
                align += (show_opcode_bytes - 1) * opcode_separator_bytes  # type: ignore[operator]
            if len(instruction.bytes) > show_opcode_bytes:
                opcodes += pwndbg.color.gray("...")
                # the length of gray("...") is 12, so we need to add extra 9 (12-3) alignment length for the invisible characters
                align += 9  # len(pwndbg.color.gray(""))
            opcodes = opcodes.ljust(align)
            if pwndbg.config.highlight_pc and i == index_of_pc:
                opcodes = ctx_color.highlight(opcodes)

        if branch_visualization:
            branch_vis_string, empty_line_branch_vis_string = create_branch_visualization_strings(
                pair_map, pair_id, maximum_pair_id, instruction.address, i == 0
            )
        else:
            branch_vis_string = None
            empty_line_branch_vis_string = ""

        # Example line:
        # ► 0x7ffff7f1aeb6 0f bd c0    <__strrchr_avx2+70>    bsr    eax, eax
        # prefix        = ►
        # address_str   = 0x555555556030
        # opcodes       = 0f bd c0                  Opcodes are enabled with the 'nearpc-num-opcode-bytes' setting
        # symbol        = <__strrchr_avx2+70>
        # asm           = bsr    eax, eax           (jump target/annotation would go here too)

        printable_elements = list(
            filter(None, (prefix, address_str, opcodes, symbol, branch_vis_string, asm))
        )

        line = " ".join(printable_elements)

        if branch_visualization and branch_vis_string:
            # Adjust the padding for the branch visualization string for the empty line
            branch_vis_padding = (
                -1  # -1 because there's a space between all printable_elements, so n-1 spaces
                + len(printable_elements)
                + len(prefix_sign)
                + addresses_max_length
                + len(opcodes)
                + symbols_max_length
            )
            empty_line_branch_vis_string = rjust_colored(
                empty_line_branch_vis_string,
                branch_vis_padding + COLUMNS_ALLOCATED_FOR_BRANCH_VISUALIZATION,
            )
        # FIXME(provider, integration): can we look into doing this on the decompiler side?
        # if show_comments:
        #     # Pull comments from integration if possible
        #     result += [
        #         " "
        #         * (len(pwndbg.color.unstylize(line)) - len(pwndbg.color.unstylize(asm).lstrip()))
        #         + c.integration_comments(x)
        #         for x in pwndbg.dintegration.provider.get_comment_lines(instr.address)
        #     ]

        # For Comment Function
        try:
            line += " " * 10 + ctx_color.comment(
                pwndbg.commands.comments.file_lists[pwndbg.aglib.proc.exe()][
                    hex(instruction.address)
                ]
            )
        except Exception:
            pass

        result.append(line)

        # For call instructions, attempt to resolve the target and
        # determine the number of arguments.
        if show_args:
            result.extend(
                f"{'':>8}{arg}" for arg in pwndbg.arguments.format_args(instruction=instruction)
            )

        # If this instruction deserves a down arrow to indicate a taken branch
        if instruction.split == SplitType.BRANCH_TAKEN:
            result.append(empty_line_branch_vis_string + c.branch_marker(f"{nearpc_branch_marker}"))

        # Otherwise if it's a branch and it *is* contiguous, just put an empty line.
        elif instruction.split == SplitType.BRANCH_NOT_TAKEN:
            if nearpc_branch_marker_contiguous:
                if empty_line_branch_vis_string:
                    result.append(empty_line_branch_vis_string)
                else:
                    result.append(f"{nearpc_branch_marker_contiguous}")

    return result


nearpc.next_pc = 0



from __future__ import annotations

import argparse

import pwndbg
import pwndbg.aglib.heap.jemalloc as jemalloc
import pwndbg.color.context as ctx_color
import pwndbg.commands
import pwndbg.dbg_mod
from pwndbg.color import message
from pwndbg.commands import CommandCategory

parser = argparse.ArgumentParser(
    description="Returns extent information for pointer address allocated by jemalloc"
)
parser.add_argument("addr", type=int, help="Address of the allocated memory location")


@pwndbg.commands.Command(parser, category=CommandCategory.JEMALLOC)
def jemalloc_find_extent(addr) -> None:
    print(ctx_color.banner("Jemalloc find extent"))
    print("This command was tested only for jemalloc 5.3.0 and does not support lower versions")
    print()

    addr = int(addr)

    try:
        rtree = jemalloc.RTree.get_rtree()
        extent = rtree.lookup_hard(addr)
        if extent is None:
            print(message.error("ERROR: Extent not found"))
            return
        # print pointer address first, then extent address then extent information
        print(f"Pointer Address: {hex(addr)}")
        print(f"Extent Address: {hex(extent.extent_address)}")
        print()

        jemalloc_extent_info(extent.extent_address, header=False)
    except pwndbg.dbg_mod.Error as e:
        print(message.error(f"ERROR: {e}"))
        return


parser = argparse.ArgumentParser(description="Prints extent information for the given address")
parser.add_argument("addr", type=int, help="Address of the extent metadata")
parser.add_argument(
    "-v", "--verbose", action="store_true", help="Print all chunk fields, even unused ones."
)


@pwndbg.commands.Command(parser, category=CommandCategory.JEMALLOC)
def jemalloc_extent_info(addr, verbose=False, header=True) -> bool:
    if header:
        print(ctx_color.banner("Jemalloc extent info"))
        print("This command was tested only for jemalloc 5.3.0 and does not support lower versions")
        print()

    try:
        extent = jemalloc.Extent(int(addr))

        print(f"Allocated Address: {hex(extent.allocated_address)}")
        print(f"Extent Address: {hex(extent.extent_address)}")

        print(f"Size: {hex(extent.size)}")
        print(f"Small class: {extent.has_slab}")

        print(f"State: {extent.state_name}")

        if verbose:
            for bit, val in extent.bitfields.items():
                print(bit, val)
    except pwndbg.dbg_mod.Error as e:
        print(message.error(f"ERROR: {e}"))
        return False
    return True


parser = argparse.ArgumentParser(description="Prints all extents information")


@pwndbg.commands.Command(parser, category=CommandCategory.JEMALLOC)
def jemalloc_heap() -> None:
    print(ctx_color.banner("Jemalloc heap"))
    print("This command was tested only for jemalloc 5.3.0 and does not support lower versions")
    print()

    try:
        rtree = jemalloc.RTree.get_rtree()
        extents = rtree.extents
        if len(extents) == 0:
            print(message.warn("No extents found"))
            return
        for extent in extents:
            # TODO: refactor so not create copies
            if not jemalloc_extent_info(extent.extent_address, header=False):
                return
            print()
    except pwndbg.dbg_mod.Error as e:
        print(message.error(f"ERROR: {e}"))
        return



"""
Generating detailed information about signals received by the debugged process.
"""

from __future__ import annotations

from collections.abc import Callable
from enum import Enum
from typing import Any
from typing import Literal

import pwndbg
import pwndbg.aglib.vmmap
import pwndbg.dbg_mod
import pwndbg.lib.arch


def get_segv_pkuerr_description() -> str | None:
    """Retrieve the long description for SEGV_PKUERR, if applicable."""
    curr_thread = pwndbg.dbg.selected_thread()
    if curr_thread is None:
        return None
    siginfo = curr_thread.siginfo()
    if siginfo is None:
        return None
    fault_mem_page = pwndbg.aglib.vmmap.find(siginfo.sigfault.si_addr)
    if fault_mem_page is None or fault_mem_page.protection_key is None:
        return None
    msg = f" Violated protection key {fault_mem_page.protection_key}"
    pkru = pwndbg.aglib.regs.read_reg("pkru")
    if pkru is None:
        return msg
    ad = 1 if pkru & (1 << (fault_mem_page.protection_key * 2)) else 0
    wd = 1 if pkru & (1 << (fault_mem_page.protection_key * 2 + 1)) else 0
    msg += f"(AD={ad}, WD={wd})"
    return msg


SIGNALS = Literal[
    "SIGABRT",
    "SIGALRM",
    "SIGBUS",
    "SIGCHLD",
    "SIGCLD",
    "SIGCONT",
    "SIGEMT",
    "SIGFPE",
    "SIGHUP",
    "SIGILL",
    "SIGINFO",
    "SIGINT",
    "SIGIO",
    "SIGIOT",
    "SIGKILL",
    "SIGLOST",
    "SIGPIPE",
    "SIGPOLL",
    "SIGPROF",
    "SIGPWR",
    "SIGQUIT",
    "SIGSEGV",
    "SIGSTKFLT",
    "SIGSTOP",
    "SIGTSTP",
    "SIGSYS",
    "SIGTERM",
    "SIGTRAP",
    "SIGTTIN",
    "SIGTTOU",
    "SIGUNUSED",
    "SIGURG",
    "SIGUSR1",
    "SIGUSR2",
    "SIGVTALRM",
    "SIGXCPU",
    "SIGXFSZ",
    "SIGWINCH",
]

COMMON_NUM_TO_SIGNAL_MAPPING: dict[int, SIGNALS] = {
    1: "SIGHUP",
    2: "SIGINT",
    3: "SIGQUIT",
    4: "SIGILL",
    5: "SIGTRAP",
    6: "SIGABRT",
    8: "SIGFPE",
    9: "SIGKILL",
    11: "SIGSEGV",
    13: "SIGPIPE",
    14: "SIGALRM",
    15: "SIGTERM",
}

X86_64_NUM_TO_SIGNAL_MAPPING: dict[int, SIGNALS] = {
    **COMMON_NUM_TO_SIGNAL_MAPPING,
    7: "SIGBUS",
    10: "SIGUSR1",
    12: "SIGUSR2",
    16: "SIGSTKFLT",
    17: "SIGCHLD",
    18: "SIGCONT",
    19: "SIGSTOP",
    20: "SIGTSTP",
    21: "SIGTTIN",
    22: "SIGTTOU",
    23: "SIGURG",
    24: "SIGXCPU",
    25: "SIGXFSZ",
    26: "SIGVTALRM",
    27: "SIGPROF",
    28: "SIGWINCH",
    29: "SIGIO",
    30: "SIGPWR",
    31: "SIGSYS",
}

PER_ARCH_SIGNAL_MAPPINGS: dict[
    pwndbg.lib.arch.PWNDBG_SUPPORTED_ARCHITECTURES_TYPE, dict[int, SIGNALS]
] = {
    "x86-64": X86_64_NUM_TO_SIGNAL_MAPPING,
}


class SegvCodeX86_64(Enum):
    SEGV_PKUERR = 4  # Protection key violation (PKU)


SHORT_SEGV_DESCRIPTIONS_X86_64: dict[SegvCodeX86_64, str] = {
    SegvCodeX86_64.SEGV_PKUERR: "SEGV_PKUERR"
}

LONG_SEGV_DESCRIPTIONS_X86_64: dict[SegvCodeX86_64, Callable[[], str | None]] = {
    SegvCodeX86_64.SEGV_PKUERR: get_segv_pkuerr_description
}

PER_ARCH_SEGV_CODES: dict[pwndbg.lib.arch.PWNDBG_SUPPORTED_ARCHITECTURES_TYPE, type[Enum]] = {
    "x86-64": SegvCodeX86_64,
}

SHORT_SEGV_DESCRIPTIONS: dict[
    pwndbg.lib.arch.PWNDBG_SUPPORTED_ARCHITECTURES_TYPE, dict[Any, str]
] = {
    "x86-64": SHORT_SEGV_DESCRIPTIONS_X86_64,
}

LONG_SEGV_DESCRIPTIONS: dict[
    pwndbg.lib.arch.PWNDBG_SUPPORTED_ARCHITECTURES_TYPE,
    dict[Any, Callable[[], str | None]],
] = {
    "x86-64": LONG_SEGV_DESCRIPTIONS_X86_64,
}


def get_segv_information() -> tuple[str, str | None]:
    """Retrieve additional information about a SIGSEGV signal, if available."""
    try:
        curr_thread = pwndbg.dbg.selected_thread()
        if curr_thread is None:
            return "SIGSEGV", None
        siginfo = curr_thread.siginfo()
        if siginfo is None:
            return "SIGSEGV", None
        si_code = siginfo.si_code
        desc_short = "SIGSEGV"
        desc_long = f" (fault address: {siginfo.sigfault.si_addr:#x})."
        curr_arch = pwndbg.aglib.arch.name
        segv_code_enum = PER_ARCH_SEGV_CODES.get(curr_arch)
        if segv_code_enum is None:
            return desc_short, desc_long

        if not any(si_code == e.value for e in segv_code_enum):
            return desc_short, desc_long

        segv_code = segv_code_enum(si_code)
        segv_type = SHORT_SEGV_DESCRIPTIONS.get(curr_arch, {}).get(segv_code, "SIGSEGV")
        if segv_type is not None:
            desc_short = segv_type
        desc_long_fn = LONG_SEGV_DESCRIPTIONS.get(curr_arch, {}).get(segv_code)
        desc_long_fn_val = desc_long_fn() if desc_long_fn else None
        if desc_long_fn_val is not None:
            desc_long += desc_long_fn_val

        return desc_short, desc_long

    except pwndbg.dbg_mod.Error:
        return "SIGSEGV", None


def get_last_signal() -> SIGNALS | None:
    """Get the last signal received by the debugged process."""
    curr_thread = pwndbg.dbg.selected_thread()
    if curr_thread is None:
        return None
    siginfo = curr_thread.siginfo()
    if siginfo is None:
        return None
    curr_arch = pwndbg.aglib.arch.name
    sig_mapping = PER_ARCH_SIGNAL_MAPPINGS.get(curr_arch, COMMON_NUM_TO_SIGNAL_MAPPING)
    return sig_mapping.get(siginfo.si_signo)



from __future__ import annotations

import argparse

import pwndbg.aglib
import pwndbg.commands
from pwndbg.commands import CommandCategory

parser = argparse.ArgumentParser(description="Modify the flags register.")
parser.add_argument("flag", type=str, help="Flag for which you want to change the value")
parser.add_argument(
    "value",
    type=int,
    help="Value to which you want to set the flag - only valid options are 0 and 1",
)


@pwndbg.commands.Command(
    parser,
    aliases=["flag"],
    category=CommandCategory.REGISTER,
    examples="""
On X86/X64:
    setflag ZF 1        -- set zero flag
    setflag CF 0        -- unset carry flag

On ARM:
    setflag Z 0         -- unset the Z cpsr/xpsr flag

To see flags registers:
    info reg eflags     -- on x86/x64
    info reg cpsr/xpsr  -- on ARM (specific register may vary)
    """,
    notes="""
This command supports flags registers that are defined for architectures in the pwndbg/regs.py file.
    """,
)
@pwndbg.commands.OnlyWhenRunning
def setflag(flag: str, value: int) -> None:
    register_set = pwndbg.aglib.regs.current

    flag = flag.upper()
    for flag_reg, flags in register_set.flags.items():
        for flag_name, bit in flags.items():
            if flag_name == flag:
                # If the size is not specified, assume it's 1
                if isinstance(bit, int):
                    size = 1
                else:
                    assert len(bit) == 2
                    size = bit[1]
                    bit = bit[0]

                max_val = (1 << size) - 1
                if value > max_val:
                    print(f"Maximum value for flag is {max_val} (size={size})")
                    return

                old_val = int(pwndbg.aglib.regs.read_reg(flag_reg))
                mask = max_val << bit
                bit_value = value << bit

                cleared_val = old_val & ~mask
                new_val = cleared_val | bit_value

                pwndbg.aglib.regs.write_reg(flag_reg, new_val)
                print(
                    f"Set flag {flag}={value} in flag register {flag_reg} (old val={old_val:#x}, new val={new_val:#x})"
                )
                return



from __future__ import annotations

import cProfile
import logging
import os
import sys
import time
import traceback

import gdb

from pwndbginit import gdbpatches  # noqa: F401
from pwndbginit.common import verify_venv


def init_logger():
    log_level_env = os.environ.get("PWNDBG_LOGLEVEL", "WARNING")
    log_level = getattr(logging, log_level_env.upper())

    root_logger = logging.getLogger()
    root_logger.setLevel(log_level)

    # Add a custom StreamHandler we will use to customize log message formatting. We
    # configure the handler later, after pwndbg has been imported.
    handler = logging.StreamHandler()
    root_logger.addHandler(handler)

    return handler


def check_doubleload():
    if "pwndbg" in sys.modules:
        print(
            "Detected double-loading of Pwndbg (likely from both .gdbinit and the Pwndbg portable build)."
        )
        print(
            "To fix this, please remove the line 'source your-path/gdbinit.py' from your .gdbinit file."
        )
        sys.exit(1)


def main() -> None:
    handler = init_logger()
    profiler = cProfile.Profile()

    start_time = None
    if os.environ.get("PWNDBG_PROFILE") == "1":
        start_time = time.time()
        profiler.enable()

    check_doubleload()
    verify_venv()

    # Force UTF-8 encoding (to_string=True to skip output appearing to the user)
    try:
        gdb.execute("set target-wide-charset UTF-8", to_string=True)
        gdb.execute("set charset UTF-8", to_string=True)
    except gdb.error as e:
        print(f"Warning: Cannot set gdb charset: '{e}'")

    import pwndbg  # noqa: F811
    import pwndbg.dbg_mod.gdb

    # Mark that pwndbg was loaded from `pwndbg` binary (for double-load detection)
    setattr(pwndbg, "_is_loaded_from_pwndbg", True)

    pwndbg.dbg = pwndbg.dbg_mod.gdb.GDB()
    pwndbg.dbg.setup()

    import pwndbg.log
    import pwndbg.profiling

    # ColorFormatter relies on pwndbg being loaded, so we can't set it up until now
    handler.setFormatter(pwndbg.log.ColorFormatter())

    pwndbg.profiling.init(profiler, start_time)
    if os.environ.get("PWNDBG_PROFILE") == "1":
        pwndbg.profiling.profiler.stop("pwndbg-load.pstats")
        pwndbg.profiling.profiler.start()

    # We need reimport it here so that it's available at the global scope
    # when some starts a Python interpreter in GDB
    gdb.execute("py import pwndbg")


def main_try():
    # We wrap everything in try/except so that we can exit GDB with an error code
    # This is used by tests to check if gdbinit.py failed
    try:
        main()
    except Exception:
        print(traceback.format_exc(), file=sys.stderr, flush=True)
        os._exit(1)



from __future__ import annotations

from collections.abc import Callable

from pwndbg.color import theme

config_status_on_color = theme.add_color_param(
    "message-status-on-color", "green", "color of on status messages"
)
config_status_off_color = theme.add_color_param(
    "message-status-off-color", "red", "color of off status messages"
)

config_notice_color = theme.add_color_param(
    "message-notice-color", "purple", "color of notice messages"
)
config_hint_color = theme.add_color_param(
    "message-hint-color", "yellow", "color of hint and marker messages"
)
config_success_color = theme.add_color_param(
    "message-success-color", "green", "color of success messages"
)
config_debug_color = theme.add_color_param("message-debug-color", "blue", "color of debug messages")
config_info_color = theme.add_color_param("message-info-color", "white", "color of info messages")
config_warning_color = theme.add_color_param(
    "message-warning-color", "yellow", "color of warning messages"
)
config_error_color = theme.add_color_param("message-error-color", "red", "color of error messages")
config_system_color = theme.add_color_param(
    "message-system-color", "light-red", "color of system messages"
)

config_exit_color = theme.add_color_param("message-exit-color", "red", "color of exit messages")
config_breakpoint_color = theme.add_color_param(
    "message-breakpoint-color", "yellow", "color of breakpoint messages"
)
config_signal_color = theme.add_color_param(
    "message-signal-color", "bold,red", "color of signal messages"
)

config_prompt_color = theme.add_color_param("prompt-color", "bold,red", "prompt color")
config_prompt_alive_color = theme.add_color_param(
    "prompt-alive-color", "bold,green", "prompt alive color"
)


def on(msg: object) -> str:
    return config_status_on_color.color_function(msg)


def off(msg: object) -> str:
    return config_status_off_color.color_function(msg)


def notice(msg: object) -> str:
    return config_notice_color.color_function(msg)


def hint(msg: object) -> str:
    return config_hint_color.color_function(msg)


def success(msg: object) -> str:
    return config_success_color.color_function(msg)


def debug(msg: object) -> str:
    return config_debug_color.color_function(msg)


def info(msg: object) -> str:
    return config_info_color.color_function(msg)


def warn(msg: object) -> str:
    return config_warning_color.color_function(msg)


def error(msg: object) -> str:
    return config_error_color.color_function(msg)


def system(msg: object) -> str:
    return config_system_color.color_function(msg)


def exit(msg: object) -> str:
    return config_exit_color.color_function(msg)


def breakpoint(msg: object) -> str:
    return config_breakpoint_color.color_function(msg)


def signal(msg: object) -> str:
    return config_signal_color.color_function(msg)


def prompt(msg: object) -> str:
    return config_prompt_color.color_function(msg)


def alive_prompt(msg: object) -> str:
    return config_prompt_alive_color.color_function(msg)


def readline_escape(func_message: Callable[[str], str], text: str) -> str:
    # For readline-based applications, non-printable escape codes must be
    # wrapped with special markers (\001 and \002). These markers inform
    # readline to ignore the escape sequences when calculating the prompt's width.
    # Without these markers, the prompt may break when navigating command history
    # with the UP arrow key or for long commands.
    return "\x01" + func_message("\x02" + text + "\x01") + "\x02"



"""
Looking up addresses for function names / symbols, and
vice-versa.
"""

from __future__ import annotations

from enum import Enum
from pathlib import Path

import gdb

from pwndbg.gdblib import gdb_version

# Symbol lookup only throws exceptions on errors, not if it failed to
# look up a symbol. We want to raise these errors so we can handle them
# properly, but there are some we haven't figured out how to fix yet, so
# we ignore those here
skipped_exceptions = (
    # This exception is being thrown by the Go typeinfo tests, we should
    # investigate why this is happening and see if we can explicitly check
    # for it with `gdb.selected_frame()`
    "No frame selected",
    # If we try to look up a TLS variable when there is no TLS, this
    # exception occurs. Ideally we should come up with a way to check for
    # this case before calling `gdb.lookup_symbol`
    "Cannot find thread-local",
)


def resolve_addr(address: int) -> str:
    """
    Retrieve the name for the symbol located at `address`
    Empty string if no symbol
    """
    # We could rewrite this function with gdb.Value+cast:
    # Or with `maint print msymbols`
    #
    # In [17]: hex(gdb.parse_and_eval('&\'malloc\''))
    # Out[17]: '0xf8f7b18065d0'
    #
    # In [18]: str(gdb.Value(0xf8f7b18065d0).cast(gdb.lookup_type("void").pointer()))
    # Out[18]: '0xf8f7b18065d0 <__GI___libc_malloc>'
    #
    # pwndbg> info symbol 0xf8f7b18065d0
    # malloc in section .text of /lib/aarch64-linux-gnu/libc.so.6
    #
    # In [19]: str(gdb.Value(0xf8f7b18065d0+0x15).cast(gdb.lookup_type("void").pointer()))
    # Out[20]: '0xf8f7b18065e5 <__GI___libc_malloc+21>'

    # Note: we do not return "" on `address < pwndbg.aglib.memory.MMAP_MIN_ADDR`
    # because this may be used to find out the symbol name on PIE binaries that weren't started yet
    # and then their symbol addresses can be found by GDB on their (non-rebased) offsets

    # Fast path: GDB's `info symbol` returns 'Numeric constant too large' here
    if address >= ((1 << 64) - 1):
        return ""

    # This sucks, but there's not a GDB API for this.
    # Workaround for a bug with Rust language, see #2094
    try:
        result = gdb.execute(f"info symbol 0x{address:x}", to_string=True, from_tty=False)
    except gdb.error:
        return ""

    if result.startswith("No symbol"):
        return ""

    # If there are newlines, which means that there are multiple symbols for the address
    # then use the first one (see also #1610)
    result = result[: result.index("\n")]

    # See https://github.com/bminor/binutils-gdb/blob/d1702fea87aa62dff7de465464097dba63cc8c0f/gdb/printcmd.c#L1594-L1624
    # The most often encountered formats looks like this:
    #   "main in section .text of /bin/bash"
    #   "main + 3 in section .text of /bin/bash"
    #   "system + 1 in section .text of /lib/x86_64-linux-gnu/libc.so.6"
    #   "No symbol matches system-1"
    # But there are some others that we have to account for as well
    if " in section " in result:
        loc_string, _ = result.split(" in section ")
    elif " in load address range of " in result:
        loc_string, _ = result.split(" in load address range of ")
    elif " overlay section " in result:
        result, _ = result.split(" overlay section ")
        loc_string, _ = result.split(" in ")
    else:
        loc_string = ""

    # If there is 'main + 87' we want to replace it with 'main+87' etc.
    return loc_string.replace(" + ", "+")


def _global_static_symbol_to_address(
    obj: gdb.Objfile | None, name: str, domain: Domain
) -> gdb.Value | None:
    try:
        # gdb.lookup_static_symbol Search order:
        # - global static in your module
        # - global static in other module
        symbol_obj = (obj or gdb).lookup_static_symbol(name, domain=DOMAIN_MAPPING[domain])
        if symbol_obj and domain.validate(symbol_obj):
            return symbol_obj.value().address
    except gdb.error:
        pass
    return None


def _global_exported_symbol_to_address(
    obj: gdb.Objfile | None, name: str, domain: Domain
) -> gdb.Value | None:
    try:
        # gdb.lookup_global_symbol Search order:
        # - global in your module
        # - global in other module
        symbol_obj = (obj or gdb).lookup_global_symbol(name, domain=DOMAIN_MAPPING[domain])
        if symbol_obj and domain.validate(symbol_obj):
            return symbol_obj.value().address
    except gdb.error:
        pass
    return None


def _frame_any_symbol_to_address(name: str, domain: Domain) -> gdb.Value | None:
    try:
        # gdb.lookup_symbol Search order:
        # - local scope
        # - global in your module
        # - global static in your module
        # - global in other module
        # - global static in other module
        symbol_obj, _ = gdb.lookup_symbol(name, domain=DOMAIN_MAPPING[domain])
        if symbol_obj and domain.validate(symbol_obj):
            if symbol_obj.needs_frame:
                return symbol_obj.value(gdb.selected_frame()).address
            return symbol_obj.value().address
    except gdb.error as e:
        if all(x not in str(e) for x in skipped_exceptions):
            raise e
    return None


def _fallback_any_symbol_to_address(name: str, global_only: bool = False) -> gdb.Value | None:
    try:
        # Unfortunately, `gdb.lookup_symbol` does not seem to handle all
        # symbols, so we need to fallback to using `gdb.parse_and_eval`. See
        # https://sourceware.org/pipermail/gdb/2022-October/050362.html
        # (We tried parsing the output of the `info address` before, but there were some issues. See #1628 and #1666)
        if "\\" in name:
            # Is it possible that happens? Probably not, but just in case
            raise ValueError(f"Symbol {name!r} contains a backslash")
        sanitized_symbol_name = name.replace("'", "\\'")

        # gdb.parse_and_eval Search order:
        # - local scope
        # - global in your module
        # - global static in your module
        # - global in other module
        # - global static in other module

        # global_context is only supported in GDB14+
        if gdb_version[0] >= 14:
            return gdb.parse_and_eval(f"&'{sanitized_symbol_name}'", global_context=global_only)  # type: ignore[call-arg]

        return gdb.parse_and_eval(f"&'{sanitized_symbol_name}'")
    except gdb.error:
        return None


class Domain(Enum):
    ANY = 1
    VARIABLE = 2
    FUNCTION = 3

    def validate(self, sym: gdb.Symbol) -> bool:
        if self == Domain.FUNCTION and gdb_version[0] < 15:
            return sym.is_function

        if self == Domain.VARIABLE:
            # For 'VARIABLE' we need manually filter out
            # We have to check for `is_function`, because TLS variables will return False in `is_variable`
            if sym.is_function:
                return False
        return True


# SYMBOL_FUNCTION_DOMAIN is supported since GDB15+
if gdb_version[0] < 15:
    gdb.SYMBOL_FUNCTION_DOMAIN = gdb.SYMBOL_VAR_DOMAIN  # type: ignore[attr-defined]

DOMAIN_MAPPING = {
    # Gdb supported types:
    # https://github.com/bminor/binutils-gdb/blob/e998ba604f8b1498c8ad43f2c19fee097b6131ef/gdb/sym-domains.def
    Domain.ANY: gdb.SYMBOL_VAR_DOMAIN,
    # gdb.SYMBOL_VAR_DOMAIN search (due to historical reasons in GDB) includes:
    # - variables
    # - function names
    # - typedef names (!sic, this is not symbol)
    # - enum type values (!sic, this is not symbol)
    # Note: This queries SYMBOL_VAR_DOMAIN, SYMBOL_TYPE_DOMAIN, and SYMBOL_FUNCTION_DOMAIN.
    Domain.VARIABLE: gdb.SYMBOL_VAR_DOMAIN,
    # Specifically for variables. Requires manual filtering to exclude other types.
    Domain.FUNCTION: gdb.SYMBOL_FUNCTION_DOMAIN,  # type: ignore[attr-defined]
    # Specifically for functions.
}

order_prefs = {
    True: (_global_static_symbol_to_address, _global_exported_symbol_to_address),
    False: (_global_exported_symbol_to_address, _global_static_symbol_to_address),
}


def lookup_symbol(
    name: str,
    *,
    prefer_static: bool = False,
    domain: Domain = Domain.ANY,
    objfile_endswith: str | None = None,
) -> gdb.Value | None:
    """
    Get the address for `symbol`
    """

    objfile: gdb.Objfile | None = None
    if objfile_endswith is not None:
        for obj in gdb.selected_inferior().progspace.objfiles():
            # FIXME: Workaround for #3641
            if str(Path(obj.filename).resolve()).endswith(objfile_endswith):
                objfile = obj
                break
        if objfile is None:
            raise gdb.GdbError(f"Objfile '{objfile_endswith}' not found")

    for func in order_prefs[prefer_static]:
        if (val := func(objfile, name, domain)) is not None:
            return val

    # FIXME: Due to a bug in GDB, some symbols (e.g., malloc / __GI___libc_malloc)
    #   may return WRONG-ADDRESS when queried.
    #   For more details, see: https://github.com/pwndbg/pwndbg/issues/2613
    #   Can be fixed by using GdbMinimalSymbols: `maint print msymbols`
    return _fallback_any_symbol_to_address(name, global_only=True)


def lookup_frame_symbol(
    name: str,
    *,
    domain: Domain = Domain.ANY,
) -> gdb.Value | None:
    """
    Get the address for local `symbol` from frame, in most time you don't need it
    """

    if (val := _frame_any_symbol_to_address(name, domain)) is not None:
        return val

    # fallback, because of bug in gdb for some symbols eg: malloc / __GI___libc_malloc
    return _fallback_any_symbol_to_address(name, global_only=False)



"""
Common exception handling.
"""

from __future__ import annotations

import functools
import sys
import traceback

import pwndbg.lib.cache
import pwndbg.lib.stdio
from pwndbg import config
from pwndbg.color import message

try:
    import ipdb as pdb
except ImportError:
    import pdb

_rich_console = None


def print_exception(exception_msg) -> None:
    global _rich_console

    if _rich_console is None:
        try:
            from rich.console import Console

            _rich_console = Console()
        except ImportError:
            _rich_console = ...

    if not isinstance(_rich_console, type(Ellipsis)):
        _rich_console.print_exception()
    else:
        print(exception_msg)


verbose = config.add_param(
    "exception-verbose",
    False,
    "print a full stacktrace for exceptions raised in Pwndbg commands",
)
debug = config.add_param(
    "exception-debugger", False, "whether to debug exceptions raised in Pwndbg commands"
)


@pwndbg.lib.cache.cache_until("forever")
def inform_report_issue(exception_msg) -> None:
    """
    Informs user that he can report an issue.
    The use of caching makes it reporting only once for a given exception message.
    """
    print(
        message.notice(
            "If that is an issue, you can report it on https://github.com/pwndbg/pwndbg/issues\n"
            "(Please don't forget to search if it hasn't been reported before)\n"
            "To generate the report and open a browser, you may run "
        )
        + message.hint("`bugreport --run-browser`")
        + message.notice("\nPS: Pull requests are welcome")
    )


def inform_verbose_and_debug() -> None:
    print(
        message.notice("For more info invoke `")
        + message.hint("set exception-verbose on")
        + message.notice("` and rerun the command\nor debug it by yourself with `")
        + message.hint("set exception-debugger on")
        + message.notice("`")
    )


def handle(name="Error") -> None:
    """Displays an exception to the user, optionally displaying a full traceback
    and spawning an interactive post-moretem debugger.

    Notes:
        - ``set exception-verbose on`` enables stack traces.
        - ``set exception-debugger on`` enables the post-mortem debugger.
    """

    # This is for unit tests so they fail on exceptions instead of displaying them.
    if getattr(sys, "_pwndbg_unittest_run", False) is True:
        E, V, T = sys.exc_info()
        e = E(V)
        e.__traceback__ = T
        raise e

    # Display the error
    if debug or verbose:
        exception_msg = traceback.format_exc()
        print_exception(exception_msg)
        inform_report_issue(exception_msg)

    else:
        exc_type, exc_value, exc_traceback = sys.exc_info()

        print(message.error(f"Exception occurred: {name}: {exc_value} ({exc_type})"))

        inform_verbose_and_debug()

    # Break into the interactive debugger
    if debug:
        with pwndbg.lib.stdio.stdio:
            pdb.post_mortem()


@functools.wraps(pdb.set_trace)
def set_trace() -> None:
    """Enable sane debugging in Pwndbg by switching to the "real" stdio."""
    debugger = pdb.Pdb(
        stdin=sys.__stdin__, stdout=sys.__stdout__, skip=["pwndbg.lib.stdio", "pwndbg.exception"]
    )
    debugger.set_trace()


pdb.set_trace = set_trace


@config.trigger(verbose, debug)
def update() -> None:
    enable = verbose or debug
    pwndbg.dbg.set_python_diagnostics(bool(enable))



from __future__ import annotations

import argparse
from typing import NamedTuple

import pwndbg.aglib.kernel
import pwndbg.color.message as message
import pwndbg.commands
from pwndbg.commands import CommandCategory


class Option(NamedTuple):
    name: str
    desired: bool = True
    cmdline_option: str = None


# TODO: Account for commandline params, module params, and sysctls

_hardening_options = [
    # Stack overflow protection
    Option("STACKPROTECTOR"),
    Option("STACKPROTECTOR_STRONG"),
    Option("SHADOW_CALL_STACK"),
    # RWX
    Option("STRICT_KERNEL_RWX"),
    Option("STRICT_MODULE_RWX"),
    Option("DEBUG_WX"),
    # ASLR
    Option("RANDOMIZE_BASE"),
    Option("RANDOMIZE_KSTACK_OFFSET_DEFAULT"),
    Option("RANDOMIZE_MODULE_REGION_FULL"),
    Option("KALLSYMS", False),
    # Memory allocation
    Option("SLAB_FREELIST_HARDENED"),
    Option("SLAB_FREELIST_RANDOM"),
    Option("KFENCE"),
    # Uninitialized data
    Option("INIT_ON_ALLOC_DEFAULT_ON"),
    Option("INIT_STACK_ALL_ZERO"),
    Option("INIT_ON_FREE_DEFAULT_ON"),
    # CFI
    Option("CFI_CLANG"),
    Option("CFI_PERMISSIVE", False),
    # Access control
    Option("SECURITY"),
    Option("SECURITY_YAMA"),
    Option("SECURITY_SELINUX_DISABLE", False),
    Option("SECURITY_SELINUX_BOOTPARAM", False),
    Option("SECURITY_SELINUX_DEVELOP", False),
    # Tracing
    Option("KPROBES", False),
    Option("FTRACE", False),
    Option("KPROBE_EVENTS", False),
    Option("UPROBE_EVENTS", False),
    Option("GENERIC_TRACER", False),
    Option("FUNCTION_TRACER", False),
    Option("STACK_TRACER", False),
    # /dev/mem
    Option("STRICT_DEVMEM"),
    Option("DEVMEM", False),
    Option("DEVKMEM", False),
    # debugfs
    Option("DEBUG_FS", False),
    Option("PTDUMP_DEBUGFS", False),
    # Misc
    Option("BUG"),
    Option("MODULES", False),
    Option("USERFAULTFD", False),
    Option("FORTIFY_SOURCE"),
    Option("STATIC_USERMODEHELPER"),
    Option("HARDENED_USERCOPY"),
    Option("RODATA_FULL_DEFAULT_ENABLED"),
    Option("RANDSTRUCT_FULL"),
    Option("TRIM_UNUSED_KSYMS"),
    Option("SECURITY_DMESG_RESTRICT"),
    Option("PROC_KCORE", False),
    Option("PROC_VMCORE", False),
    Option("COMPAT_VDSO", False),
    Option("BINFMT_MISC", False),
]

_x86_hardening_options = [
    Option("X86_SMAP"),
    Option("IA32_EMULATION", False),
    Option("X86_X32", False),
]

_arch_hardening_options = {}
_arch_hardening_options["i386"] = _x86_hardening_options
_arch_hardening_options["x86-64"] = _x86_hardening_options
_arch_hardening_options["aarch64"] = [
    Option("ARM64_PAN"),
    Option("ARM64_EPAN"),
    Option("ARM64_SW_TTBR0_PAN"),
    Option("ARM64_PTR_AUTH"),
    Option("ARM64_PTR_AUTH_KERNEL"),
    Option("ARM64_BTI"),
    Option("ARM64_BTI_KERNEL"),
    Option("ARM64_MTE"),
]

parser = argparse.ArgumentParser(description="Checks for kernel hardening configuration options.")


@pwndbg.commands.Command(parser, category=CommandCategory.KERNEL)
@pwndbg.commands.OnlyWhenQemuKernel
@pwndbg.commands.OnlyWhenPagingEnabled
def kchecksec() -> None:
    kconfig = pwndbg.aglib.kernel.kconfig()

    options = _hardening_options + _arch_hardening_options.get(pwndbg.aglib.arch.name, [])
    for opt in options:
        config_name = opt.name
        val = kconfig.get(config_name)
        color_func = message.error
        if (opt.desired and val) or (not opt.desired and not val):
            color_func = message.success

        if val:
            print(color_func(f"CONFIG_{config_name} = {val}"))
        else:
            print(color_func(f"CONFIG_{config_name} not set"))



from __future__ import annotations

import argparse

from rich.console import Console
from rich.table import Table

import pwndbg
import pwndbg.aglib.commpage
import pwndbg.commands
from pwndbg.commands import CommandCategory


def pretty_bytes(data: bytes) -> str:
    if not data:
        return "b''"

    if len(set(data)) == 1 and len(data) > 1:
        return f"b'\\x{data[0]:02x}' <repeats {len(data)} times>"

    return repr(data)


parser = argparse.ArgumentParser(description="Dumps all values from the macOS commpage.")
parser.add_argument("-v", "--verbose", action="store_true", help="Print detailed information.")


@pwndbg.commands.Command(parser, category=CommandCategory.DARWIN)
def commpage(verbose: bool = False):
    table = Table()

    table.add_column("Name", style="cyan", no_wrap=True)
    table.add_column("Address", style="magenta", no_wrap=True)
    table.add_column("Type", style="magenta", no_wrap=True)
    table.add_column("Value", style="green", overflow="fold")
    if verbose:
        table.add_column("Description", style="yellow", overflow="fold")

    for comm in pwndbg.aglib.commpage.get_commpage_fields():
        val = comm.unpack()
        style = None
        if comm.is_undocumented():
            style = "red"
        if comm.is_unused():
            style = "yellow"

        if isinstance(val, bytes):
            val = pretty_bytes(val)

        rows = [comm.name, hex(comm.real_addr()), comm.ctype, str(val)]
        if verbose:
            rows.append(comm.desc)
        table.add_row(*rows, style=style)

    Console().print(table)



from __future__ import annotations

import argparse
import os
import tempfile

import niche_elf
import niche_elf.datatypes

import pwndbg
import pwndbg.aglib.kernel.kallsyms
import pwndbg.commands
import pwndbg.dbg_mod
from pwndbg.color import message
from pwndbg.commands import CommandCategory

parser = argparse.ArgumentParser(description="Lookup kernel symbols")

parser.add_argument(
    "symbol", type=str, nargs="?", default="", help="Address or symbol name to lookup"
)
parser.add_argument(
    "-a", "--apply", action="store_true", help="applies all the symbols that satisfy the filter"
)


@pwndbg.commands.Command(
    parser,
    aliases=["kallsyms", "ks"],
    category=CommandCategory.KERNEL,
    notes="""
Using `--apply` makes sense for kernel modules. If you want to symbolize the whole kernel,
use vmlinux-to-elf (https://github.com/marin-m/vmlinux-to-elf) or compile it yourself.
""",
)
@pwndbg.commands.OnlyWhenQemuKernel
@pwndbg.commands.OnlyWhenPagingEnabled
def klookup(symbol: str, apply: bool) -> None:
    ksyms = pwndbg.aglib.kernel.kallsyms.get()
    syms = []
    try:
        symbol_addr = int(symbol, 0)
        for sym in ksyms:
            if sym[2] == symbol_addr:
                syms.append(sym)
        if len(syms) == 0:
            print(message.error(f"No symbol found at {symbol_addr:#x}"))
    except (ValueError, TypeError):
        for sym in ksyms:
            if symbol is None or symbol in sym[0]:
                syms.append(sym)
        if len(syms) == 0:
            print(message.error(f"No symbol found for {symbol}"))

    if not (apply and symbol == ""):
        for sym_name, sym_type, sym_addr in syms:
            print(message.success(f"{sym_addr:#x} {sym_type} {sym_name}"))

    if apply:
        if pwndbg.dbg.name() == pwndbg.dbg_mod.DebuggerType.LLDB:
            print(message.error("Symbolication is not yet supported on LLDB."))
            # Until we implement add_symbol_file for LLDB.
            return

        base: int | None = pwndbg.aglib.kernel.kbase()

        if base is None:
            # I would be suprised if this was actually possible, we managed to find kallsyms but not
            # kbase?
            # But anyway, passing 0 to ELFFile and no ADDR argument to add_symbol_file should still work.
            elf: niche_elf.ELFFile = niche_elf.ELFFile(0)
        else:
            elf = niche_elf.ELFFile(base)

        for sym_name, sym_type, sym_addr in syms:
            # I trust bata: bata24/gef.py:create_symboled_elf()
            if sym_type and sym_type in "abcdefghijklmnopqrstuvwxyz":
                bind: int = niche_elf.datatypes.Constants.STB_LOCAL
            else:
                bind = niche_elf.datatypes.Constants.STB_GLOBAL

            if sym_type in ["T", "t", "W", None]:
                elf.add_function(sym_name, sym_addr, bind=bind)
            else:
                elf.add_object(sym_name, sym_addr, bind=bind)

        _, elf_path = tempfile.mkstemp(prefix="ks-symbols-", suffix=".elf")
        elf.write(elf_path)

        pwndbg.dbg.selected_inferior().add_symbol_file(elf_path, base)

        print(message.success(f"Added {len(syms)} symbols"))

        # Delete the file after GDB closes its file descriptor.
        os.unlink(elf_path)
        return



from __future__ import annotations

import argparse

import pwndbg.aglib.nearpc
import pwndbg.aglib.symbol
import pwndbg.commands
import pwndbg.dbg_mod
from pwndbg.commands import CommandCategory

nearpc_lines = pwndbg.config.add_param(
    "nearpc-lines", 10, "number of lines to print for the nearpc command"
)

nearpc_backwards_lines = pwndbg.config.add_param(
    "nearpc-backwards-lines", 5, "number of lines before the pc to print for the nearpc command"
)

parser = argparse.ArgumentParser(description="Disassemble near a specified address.")
parser.add_argument(
    "pc",
    type=int,
    nargs="?",
    default=None,
    help="Address to disassemble near. If this is the only argument and the value provided is small enough, it is interpreted as lines instead.",
)
parser.add_argument(
    "lines",
    type=int,
    nargs="?",
    default=None,
    help="Number of lines to disassemble.",
)
parser.add_argument(
    "-r",
    "--reverse",
    type=int,
    help="Number of lines to show before the address.",
)
parser.add_argument(
    "-t",
    "--total",
    type=int,
    help="Total number of lines to show. This results in dynamic number of forward instructions depending on how many cached instructions are used.",
)
parser.add_argument(
    "-e",
    "--emulate",
    action="store_true",
    help="Whether to emulate instructions to find the next ones or just linearly disassemble.",
)

parser.add_argument(
    "-n",
    "--no-branch",
    action="store_true",
    help="Disable showing branch visualizations.",
)

parser.add_argument(
    "-f",
    "--function",
    type=int,
    default=None,
    help="Disassemble an entire function. Takes an expression (such as a function name or address) and disassembles the function surrounding the evaluated address.",
)


@pwndbg.commands.Command(parser, aliases=["pdisass", "u"], category=CommandCategory.DISASS)
@pwndbg.commands.OnlyWhenRunning
def nearpc(
    pc=None,
    lines=None,
    reverse=None,
    total=None,
    emulate=False,
    use_cache=False,
    linear=True,
    no_branch=False,
    function=None,
) -> None:
    """
    Disassemble near a specified address.
    """
    # nearpc is flexible in the first argument (it can be an address or the number of lines to disassemble).
    # Save the first argument, which depending on the context might be the explicitly requested number of lines to disassemble.
    # None if not provided
    first_input_argument = pc

    # Fix the case where we only have one argument, and
    # it's a small value.
    if lines is None and (pc is not None and int(pc) < 0x100):
        lines = pc
        pc = None

    if pc is None:
        pc = pwndbg.aglib.regs.pc

    if lines is None:
        lines = int(nearpc_lines)

    back_lines = 0

    if reverse is None and total is None:
        back_lines = min(int(nearpc_backwards_lines), lines - 1)
    elif reverse is not None:
        back_lines = reverse
    elif total is not None:
        # -t was specified
        back_lines = min(int(nearpc_backwards_lines), total - 1)

    end_address = None
    if function is not None:
        # Emulate GDB behavior of "disass" - it disassembles the entire function in which
        # the input address resides. User can input integer or string name of function, or an expression
        boundaries = pwndbg.aglib.symbol.resolve_function_boundaries(function)
        if boundaries is None:
            print(f"Error: function boundaries of '{function}' could not be found")
            return
        pc, end_address = boundaries

        if end_address < pc:
            print(f"Error: function boundaries  of '{function}' could not be found")
            return

        if end_address - pc > 0x1000:
            print(
                f"Warning: detected very long function of length {hex(end_address - pc)} bytes. This may block for a while."
            )

        if first_input_argument is None:
            # If user didn't provide a minimum bound on number of instructions, make
            # sure we choose a number large enough to disassemble the entire function
            lines = end_address - pc
        back_lines = 0

    print(
        "\n".join(
            pwndbg.aglib.nearpc.nearpc(
                pc=pc,
                lines=lines,
                back_lines=back_lines,
                total_lines=total,
                emulate=emulate,
                repeat=nearpc.repeat,
                use_cache=use_cache,
                linear=linear,
                branch_visualization=not no_branch,
                end_address=end_address,
            )
        )
    )


parser = argparse.ArgumentParser(
    description="Like nearpc, but will emulate instructions from the current $PC forward."
)
parser.add_argument("pc", type=int, nargs="?", default=None, help="Address to emulate near.")
parser.add_argument(
    "lines",
    type=int,
    nargs="?",
    default=None,
    help="Number of lines to disassemble.",
)

parser.add_argument(
    "-r",
    "--reverse",
    type=int,
    help="Number of lines to show before the address.",
)

parser.add_argument(
    "-t",
    "--total",
    type=int,
    help="Total number of lines to show. This results in dynamic number of forward instructions depending on how many cached instructions are used.",
)


@pwndbg.commands.Command(parser, category=CommandCategory.DISASS)
@pwndbg.commands.OnlyWhenRunning
def emulate(pc=None, lines=None, reverse=None, total=None, emulate_=True) -> None:
    """
    Like nearpc, but will emulate instructions from the current $PC forward.
    """
    nearpc.repeat = emulate.repeat
    nearpc(
        pc=pc,
        lines=lines,
        reverse=reverse,
        total=total,
        emulate=emulate_,
        use_cache=True,
        linear=False,
        no_branch=True,
    )



from __future__ import annotations

from typing import Any
from typing import TypeAlias

from pycparser import CParser  # type: ignore # noqa: PGH003
from pycparser import c_ast

from pwndbg.lib.functions import Argument
from pwndbg.lib.functions import Function

CAstNode: TypeAlias = (
    c_ast.EllipsisParam
    | c_ast.PtrDecl
    | c_ast.ArrayDecl
    | c_ast.FuncDecl
    | c_ast.Struct
    | c_ast.Union
    | c_ast.Enum
)


def extractTypeAndName(n: CAstNode, defaultName: str | None = None) -> tuple[str, int, str] | None:
    if isinstance(n, c_ast.EllipsisParam):
        return ("int", 0, "vararg")

    t = n.type
    d = 0

    while isinstance(t, (c_ast.PtrDecl, c_ast.ArrayDecl)):
        d += 1
        children = dict(t.children())
        t = children["type"]

    if isinstance(t, c_ast.FuncDecl):
        return extractTypeAndName(t)

    if isinstance(t.type, (c_ast.Struct, c_ast.Union, c_ast.Enum)):
        typename = t.type.name
    else:
        typename = t.type.names[0]

    if typename == "void" and d == 0 and not t.declname:
        return None

    name = t.declname or defaultName or ""
    return typename.lstrip("_"), d, name.lstrip("_")


def Stringify(X: Function | Argument) -> str:
    return f"{X.type} {X.derefcnt * '*'} {X.name}"


def ExtractFuncDecl(node: CAstNode, verbose: bool = False) -> Function | None:
    # The function name needs to be dereferenced.
    result = extractTypeAndName(node)
    if result is None:
        return None

    ftype, fderef, fname = result

    if not fname:
        print("Skipping function without a name!")
        print(node.show())
        return None

    fargs: list[Argument] = []
    for i, (argName, arg) in enumerate(node.args.children()):
        defname = f"arg{i}"
        argdata = extractTypeAndName(arg, defname)
        if argdata is not None:
            a = Argument(*argdata)
            fargs.append(a)

    Func = Function(ftype, fderef, fname, fargs)

    if verbose:
        print(Stringify(Func) + "(" + ",".join(Stringify(a) for a in Func.args) + ");")

    return Func


def ExtractAllFuncDecls(ast: CAstNode, verbose: bool = False):
    Functions: dict[str, Function] = {}

    class FuncDefVisitor(c_ast.NodeVisitor):
        def visit_FuncDecl(self, node: CAstNode, *a: Any) -> None:
            f = ExtractFuncDecl(node, verbose)
            Functions[f.name] = f

    FuncDefVisitor().visit(ast)

    return Functions


def ExtractFuncDeclFromSource(source: str) -> Function | None:
    try:
        p = CParser()
        ast: CAstNode = p.parse(source + ";")
        funcs = ExtractAllFuncDecls(ast)
        for name, func in funcs.items():
            return func
    except Exception:
        import traceback

        traceback.print_exc()
        # eat it
    return None



"""
Provides values which would be available from /proc which
are not fulfilled by other modules and some process/gdb flow
related information.
"""

from __future__ import annotations

import functools
from collections.abc import Callable
from typing import TypeVar

from elftools.elf.relocation import Relocation
from typing_extensions import ParamSpec

import pwndbg
import pwndbg.aglib.qemu
import pwndbg.lib.arch
import pwndbg.lib.cache
import pwndbg.lib.memory

P = ParamSpec("P")
T = TypeVar("T")


def pid() -> int:
    return pwndbg.dbg.selected_inferior().pid()


def tid() -> int:
    return pwndbg.dbg.selected_thread().ptid()


def thread_id() -> int:
    return pwndbg.dbg.selected_thread().index()


def alive() -> bool:
    """
    Informs whether the process has a thread. However, note that it will
    still return True for a segfaulted thread. To detect that, consider
    using the `stopped_with_signal` method.
    """
    return pwndbg.dbg.selected_inferior().alive()


def is_core_file() -> bool:
    """
    Returns whether the loaded program is a corefile
    """
    return pwndbg.dbg.selected_inferior().is_core_file()


def stopped_with_signal() -> bool:
    """
    Returns whether the program has stopped with a signal

    Can be used to detect segfaults (but will also detect other signals)
    """
    return pwndbg.dbg.selected_inferior().stopped_with_signal()


@pwndbg.lib.cache.cache_until("objfile")
def exe() -> str | None:
    """
    Returns the executed file path.

    On remote targets, this path may not exist locally.

    If you need the locally referenced file use:
        `gdb.current_process().filename`
    """

    return pwndbg.dbg.selected_inferior().main_module_name()


@pwndbg.lib.cache.cache_until("start", "stop")
def binary_base_addr() -> int:
    return binary_vmmap()[0].start


@pwndbg.lib.cache.cache_until("start", "stop")
def binary_vmmap() -> tuple[pwndbg.lib.memory.Page, ...]:
    import pwndbg.aglib.vmmap

    return tuple(p for p in pwndbg.aglib.vmmap.get() if p.objfile == exe())


@pwndbg.lib.cache.cache_until("start", "objfile")
def dump_elf_data_section() -> tuple[int, int, bytes] | None:
    """
    Dump .data section of current process's ELF file
    """
    import pwndbg.aglib.elf

    return pwndbg.aglib.elf.section_by_name(exe(), ".data", try_local_path=True)


@pwndbg.lib.cache.cache_until("start", "objfile")
def dump_relocations_by_section_name(section_name: str) -> tuple[Relocation, ...] | None:
    """
    Dump relocations of a section by section name of current process's ELF file
    """
    import pwndbg.aglib.elf

    return pwndbg.aglib.elf.relocations_by_section_name(exe(), section_name, try_local_path=True)


@pwndbg.lib.cache.cache_until("start", "objfile")
def get_section_address_by_name(section_name: str) -> int:
    """
    Find section address of current process by section name
    """
    selected = pwndbg.dbg.selected_inferior()
    main = selected.main_module_name()

    for address, size, section, module in selected.module_section_locations():
        if module != main:
            continue
        if section == section_name:
            return address

    return 0


def OnlyWhenRunning(func: Callable[P, T]) -> Callable[P, T | None]:
    @functools.wraps(func)
    def wrapper(*a: P.args, **kw: P.kwargs) -> T | None:
        if alive():
            return func(*a, **kw)
        return None

    return wrapper


def OnlyWhenQemuKernel(func: Callable[P, T]) -> Callable[P, T | None]:
    @functools.wraps(func)
    def wrapper(*a: P.args, **kw: P.kwargs) -> T | None:
        if pwndbg.aglib.qemu.is_qemu_kernel():
            return func(*a, **kw)
        return None

    return wrapper


def OnlyWithArch(arch_names: list[str]) -> Callable[[Callable[P, T]], Callable[P, T | None]]:
    """Decorates function to work only with the specified archictectures."""
    for arch in arch_names:
        if arch not in pwndbg.lib.arch.PWNDBG_SUPPORTED_ARCHITECTURES:
            raise ValueError(
                f"OnlyWithArch used with unsupported arch={arch}. Must be one of {', '.join(arch_names)}"
            )

    def decorator(function: Callable[P, T]) -> Callable[P, T | None]:
        @functools.wraps(function)
        def _OnlyWithArch(*a: P.args, **kw: P.kwargs) -> T | None:
            if pwndbg.aglib.arch.name in arch_names:
                return function(*a, **kw)
            return None

        return _OnlyWithArch

    return decorator



from __future__ import annotations

from pwndbg.color import theme
from pwndbg.lib.regs import BitFlags

config_prefix_color = theme.add_color_param(
    "code-prefix-color", "none", "color for 'context code' command (prefix marker)"
)
config_highlight_color = theme.add_color_param(
    "highlight-color", "green,bold", "color added to highlights like source/pc"
)
config_register_color = theme.add_color_param(
    "context-register-color", "bold", "color for registers label"
)
config_flag_value_color = theme.add_color_param(
    "context-flag-value-color", "none", "color for flags register (register value)"
)
config_flag_bracket_color = theme.add_color_param(
    "context-flag-bracket-color", "none", "color for flags register (bracket)"
)
config_flag_set_color = theme.add_color_param(
    "context-flag-set-color", "green,bold", "color for flags register (flag set)"
)
config_flag_unset_color = theme.add_color_param(
    "context-flag-unset-color", "red", "color for flags register (flag unset)"
)
config_flag_changed_color = theme.add_color_param(
    "context-flag-changed-color", "underline", "color for flags register (flag changed)"
)
config_banner_color = theme.add_color_param("banner-color", "blue", "color for banner line")
config_banner_title = theme.add_color_param("banner-title-color", "none", "color for banner title")
config_register_changed_color = theme.add_color_param(
    "context-register-changed-color", "red", "color for registers label (change marker)"
)
config_register_changed_marker = theme.add_param(
    "context-register-changed-marker", "*", "change marker for registers label"
)
config_comment = theme.add_color_param("comment-color", "gray", "color for comment")


def prefix(x: object) -> str:
    return config_prefix_color.color_function(x)


def highlight(x: object) -> str:
    return config_highlight_color.color_function(x)


def register(x: object) -> str:
    return config_register_color.color_function(x)


def register_changed(x: object) -> str:
    return config_register_changed_color.color_function(x)


def flag_bracket(x: object) -> str:
    return config_flag_bracket_color.color_function(x)


def flag_value(x: object) -> str:
    return config_flag_value_color.color_function(x)


def flag_set(x: object) -> str:
    return config_flag_set_color.color_function(x)


def flag_unset(x: object) -> str:
    return config_flag_unset_color.color_function(x)


def flag_changed(x: object) -> str:
    return config_flag_changed_color.color_function(x)


def banner(x: object) -> str:
    return config_banner_color.color_function(x)


def banner_title(x: object) -> str:
    return config_banner_title.color_function(x)


def comment(x: object) -> str:
    return config_comment.color_function(x)


def format_flags(value: int | None, flags: BitFlags, last: int | None = None):
    if value is None:
        return "<unavailable>"

    desc = flag_value(f"{value:#x}")
    if not flags:
        return desc

    names: list[str] = []
    for name, bit in flags.items():
        # If the size is not specified, assume it's 1
        if isinstance(bit, int):
            size = 1
        else:
            assert len(bit) == 2
            size = bit[1] - bit[0] + 1
            bit = bit[0]

        mask = (1 << size) - 1
        flag_val = (value >> bit) & mask

        # If the bitfield is larger than a single bit, we can't communicate the value
        # with just the case of the name, so append the actual value
        if size > 1:
            if size == 2:
                name = f"{name}:{flag_val:0{size}b}"
            else:
                name = f"{name}:{flag_val:#x}"

        if flag_val == 0:
            name = flag_unset(name.lower())
        else:
            name = flag_set(name.upper())

        if last is not None and flag_val != (last >> bit) & mask:
            name = flag_changed(name)
        names.append(name)

    return f"{desc} {flag_bracket('[')} {' '.join(names)} {flag_bracket(']')}"



from __future__ import annotations

import argparse

import pwndbg.aglib
import pwndbg.aglib.argv
import pwndbg.commands
import pwndbg.commands.telescope
from pwndbg.commands import CommandCategory


@pwndbg.commands.Command("Prints out the number of arguments.", category=CommandCategory.LINUX)
@pwndbg.commands.OnlyWhenRunning
def argc() -> None:
    print(pwndbg.aglib.argv.argc())


parser = argparse.ArgumentParser(description="Prints out the contents of argv.")
parser.add_argument(
    "i", nargs="?", type=int, default=None, help="Index of the argument to print out."
)


@pwndbg.commands.Command(parser, category=CommandCategory.LINUX)
@pwndbg.commands.OnlyWhenRunning
def argv(i: int = None) -> None:
    if i is not None:
        val = pwndbg.aglib.argv.argv(i)
        if val is None:
            print("Argv not found")
            return

        pwndbg.commands.telescope.telescope(int(val.address), 1)
        return

    start = int(pwndbg.aglib.argv.argv(0).address)
    n = pwndbg.aglib.argv.argc() + 1
    pwndbg.commands.telescope.telescope(start, n)


parser = argparse.ArgumentParser(description="Prints out the contents of the environment.")
parser.add_argument(
    "name", nargs="?", type=str, default=None, help="Name of the environment variable to see."
)


@pwndbg.commands.Command(parser, aliases=["env", "environ"], category=CommandCategory.LINUX)
@pwndbg.commands.OnlyWhenRunning
@pwndbg.commands.OnlyWhenUserspace
def envp(name: str = None):
    """
    Prints out the contents of the environment.
    """
    if name is not None:
        val = pwndbg.aglib.argv.environ(name)
        if val is None:
            print("Environ not found")
            return

        pwndbg.commands.telescope.telescope(int(val.address), 1)
        return

    start = int(pwndbg.aglib.argv.envp(0).address)
    n = pwndbg.aglib.argv.envc() + 1
    pwndbg.commands.telescope.telescope(start, n)



from __future__ import annotations

import argparse

import pwndbg.aglib.asm
import pwndbg.aglib.memory
import pwndbg.color.context
import pwndbg.color.memory
import pwndbg.color.syntax_highlight
import pwndbg.commands
import pwndbg.lib.cache
from pwndbg.aglib.disasm.disassembly import get_disassembler
from pwndbg.color import message
from pwndbg.commands import CommandCategory

# Keep old patches made so we can revert them
patches: dict[int, tuple[bytes, bytes]] = {}


parser = argparse.ArgumentParser(description="Patches given instruction with given code or bytes.")
parser.add_argument("address", type=int, help="The address to patch")
parser.add_argument("ins", type=str, help="instruction[s]")
parser.add_argument("-q", "--quiet", action="store_true", help="don't print anything")


@pwndbg.commands.Command(parser, category=CommandCategory.MISC)
@pwndbg.commands.OnlyWhenRunning
def patch(address: int, ins: str, quiet: bool) -> None:
    new_mem = pwndbg.aglib.asm.asm(ins)

    old_mem = pwndbg.aglib.memory.read(address, len(new_mem))

    patches[address] = (old_mem, new_mem)

    pwndbg.aglib.memory.write(address, new_mem)

    pwndbg.lib.cache.clear_caches()

    if not quiet:
        colored_addr = pwndbg.color.memory.get(address)
        print(f"Patched {len(new_mem)} bytes at {colored_addr}")


parser2 = argparse.ArgumentParser(description="Revert patch at given address.")
parser2.add_argument("address", type=int, help="Address to revert patch on")


@pwndbg.commands.Command(parser2, category=CommandCategory.MISC)
@pwndbg.commands.OnlyWhenRunning
def patch_revert(address: int) -> None:
    if not patches:
        print(message.notice("No patches to revert"))
        return

    if address == -1:
        for addr, (old, _new) in patches.items():
            pwndbg.aglib.memory.write(addr, old)
            print(message.notice(f"Reverted patch at {addr:#x}"))
        patches.clear()
    elif address in patches:
        old, _new = patches.pop(address)
        pwndbg.aglib.memory.write(address, old)
        print(message.notice(f"Reverted patch at {address:#x}"))
    else:
        print(message.error(f"Address {address:#x} not found in patch list"))

    pwndbg.lib.cache.clear_caches()


parser3 = argparse.ArgumentParser(description="List all patches.")


@pwndbg.commands.Command(parser3, category=CommandCategory.MISC)
@pwndbg.commands.OnlyWhenRunning
def patch_list() -> None:
    if not patches:
        print(message.hint("No patches to list"))
        return

    print(pwndbg.color.context.banner("Patches:"))
    for addr, (old, new) in patches.items():
        cs = get_disassembler(pwndbg.aglib.arch.get_capstone_constants(addr))

        old_insns = "\n".join(
            [f"{x.mnemonic} {x.op_str}".strip() for x in cs.disasm(old, offset=addr)]
        )
        new_insns = "\n".join(
            [f"{x.mnemonic} {x.op_str}".strip() for x in cs.disasm(new, offset=addr)]
        )

        colored_addr = pwndbg.color.memory.get(addr)

        old_insns, new_insns = map(
            pwndbg.color.syntax_highlight.syntax_highlight, (old_insns, new_insns)
        )

        print(
            message.hint(f"{colored_addr} ({len(new)} bytes)"),
            message.hint("\n  from:"),
            message.warn(" ".join(old_insns.replace("\n", "; ").split())),
            message.hint("\n  to  :"),
            message.warn(" ".join(new_insns.replace("\n", "; ").split())),
        )



from __future__ import annotations

import struct
from typing import Literal

import pwnlib
import pwnlib.context
from capstone6pwndbg import CS_ARCH_AARCH64
from capstone6pwndbg import CS_ARCH_ARM
from capstone6pwndbg import CS_ARCH_LOONGARCH
from capstone6pwndbg import CS_ARCH_MIPS
from capstone6pwndbg import CS_ARCH_PPC
from capstone6pwndbg import CS_ARCH_RISCV
from capstone6pwndbg import CS_ARCH_SPARC
from capstone6pwndbg import CS_ARCH_SYSTEMZ
from capstone6pwndbg import CS_ARCH_X86
from capstone6pwndbg import CS_MODE_16
from capstone6pwndbg import CS_MODE_32
from capstone6pwndbg import CS_MODE_64
from capstone6pwndbg import CS_MODE_ARM
from capstone6pwndbg import CS_MODE_BIG_ENDIAN
from capstone6pwndbg import CS_MODE_LITTLE_ENDIAN
from capstone6pwndbg import CS_MODE_LOONGARCH64
from capstone6pwndbg import CS_MODE_MCLASS
from capstone6pwndbg import CS_MODE_MIPS32
from capstone6pwndbg import CS_MODE_MIPS64
from capstone6pwndbg import CS_MODE_RISCV32
from capstone6pwndbg import CS_MODE_RISCV64
from capstone6pwndbg import CS_MODE_RISCVC
from capstone6pwndbg import CS_MODE_THUMB
from capstone6pwndbg import CS_MODE_V9
from typing_extensions import override

import pwndbg
import pwndbg.aglib
import pwndbg.aglib.disasm
import pwndbg.dbg_mod
from pwndbg.aglib import typeinfo
from pwndbg.lib.abi import ABI
from pwndbg.lib.abi import DEFAULT_ABIS
from pwndbg.lib.abi import SIGRETURN_ABIS
from pwndbg.lib.abi import SYSCALL_ABIS
from pwndbg.lib.abi import SyscallABI
from pwndbg.lib.arch import PWNDBG_SUPPORTED_ARCHITECTURES_TYPE
from pwndbg.lib.arch import PWNLIB_ARCH_MAPPINGS
from pwndbg.lib.arch import PWNLIB_PLATFORM_MAPPINGS
from pwndbg.lib.arch import ArchAttribute
from pwndbg.lib.arch import ArchDefinition
from pwndbg.lib.arch import Platform

EndianType = Literal["little", "big"]

FMT_LITTLE_ENDIAN = {1: "B", 2: "<H", 4: "<I", 8: "<Q"}
FMT_BIG_ENDIAN = {1: "B", 2: ">H", 4: ">I", 8: ">Q"}


registered_architectures: dict[PWNDBG_SUPPORTED_ARCHITECTURES_TYPE, PwndbgArchitecture] = {}


def register_arch(arch: PwndbgArchitecture):
    registered_architectures[arch.name] = arch


def get_pwndbg_architecture(name: PWNDBG_SUPPORTED_ARCHITECTURES_TYPE) -> PwndbgArchitecture | None:
    if name not in registered_architectures:
        return None

    return registered_architectures[name]


CAPSTONE_ENDIAN_MAPPING: dict[EndianType, int] = {
    "little": CS_MODE_LITTLE_ENDIAN,
    "big": CS_MODE_BIG_ENDIAN,
}


class PwndbgArchitecture(ArchDefinition):
    """
    This class defines the context of the currently debugged architecture as well as other related information of the platform.

    This includes the following information:
    - Capstone/Unicorn constants
    - ABI information
    """

    ### All subclasses must provide values for the following attributes

    max_instruction_size: int
    instruction_alignment: int
    constant_instruction_size: bool = False

    ###

    name: PWNDBG_SUPPORTED_ARCHITECTURES_TYPE
    endian: EndianType
    ptrsize: int
    """Pointer size in bytes"""
    ptrbits: int
    """Pointer size in bits"""
    ptrmask: int
    function_abi: ABI | None
    syscall_abi: SyscallABI | None
    sigreturn_abi: SyscallABI | None
    platform: Platform
    attributes: list[ArchAttribute]

    fmts: dict[int, str]
    fmt: str

    def __init__(self, name: PWNDBG_SUPPORTED_ARCHITECTURES_TYPE) -> None:
        self.name: PWNDBG_SUPPORTED_ARCHITECTURES_TYPE = name

        # We have to set some values by default
        # These will be set again by the code that detects the global architecture
        self.update(
            ArchDefinition(
                name=name,
                ptrsize=typeinfo.ptrsize,
                endian="little",
                platform=Platform.LINUX,
            )
        )

    def update(self, arch_definition: ArchDefinition) -> None:
        """
        While debugging a process, certain aspects of the architecture can change.

        For example:
        - Some architectures can change endianness dynamically.
        """
        self.platform = arch_definition.platform
        self.attributes = arch_definition.attributes

        self.endian: EndianType = arch_definition.endian

        self.ptrsize: int = arch_definition.ptrsize
        self.ptrbits: int = self.ptrsize * 8
        self.ptrmask: int = (1 << self.ptrbits) - 1

        default_abi_identifer = (self.ptrbits, self.name, "linux")

        self.function_abi = DEFAULT_ABIS.get(default_abi_identifer)
        self.syscall_abi = SYSCALL_ABIS.get(default_abi_identifer)
        self.sigreturn_abi = SIGRETURN_ABIS.get(default_abi_identifer)

        self.fmts: dict[int, str] = FMT_LITTLE_ENDIAN if self.endian == "little" else FMT_BIG_ENDIAN
        self.fmt: str = self.fmts[self.ptrsize]

    def pack(self, integer: int) -> bytes:
        return struct.pack(self.fmt, integer & self.ptrmask)

    def unpack(self, data: bytes | bytearray) -> int:
        return struct.unpack(self.fmt, data)[0]

    def pack_size(self, integer: int, size: int) -> bytes:
        return struct.pack(self.fmts[size], integer & self.ptrmask)

    def unpack_size(self, data: bytes | bytearray, size: int) -> int:
        return struct.unpack(self.fmts[size], data)[0]

    def get_capstone_constants(self, address: int) -> tuple[int, int] | None:
        """
        Return tuple of (CAPSTONE ARCH, CAPSTONE MODE) used to instantiate the Capstone disassembler for this architecture.
        """
        return None

    def get_capstone_endianness(self) -> int:
        return CAPSTONE_ENDIAN_MAPPING[self.endian]

    def read_thumb_bit(self) -> Literal[0, 1, None]:
        """
        Return 0 or 1, representing the status of the Thumb bit in the current Arm architecture

        Return None if the Thumb bit is not relevent to the current architecture
        """
        return None

    def unsigned(self, val: int, shift: int = 0) -> int:
        return (val << shift) & ((1 << self.ptrbits) - 1)


class AMD64Arch(PwndbgArchitecture):
    max_instruction_size = 16
    instruction_alignment = 1

    def __init__(self) -> None:
        super().__init__("x86-64")

    @override
    def get_capstone_constants(self, address: int) -> tuple[int, int]:
        return (CS_ARCH_X86, CS_MODE_64)


class i386Arch(PwndbgArchitecture):
    """
    32-bit mode x86
    """

    max_instruction_size = 16
    instruction_alignment = 1

    def __init__(self) -> None:
        super().__init__("i386")

    @override
    def get_capstone_constants(self, address: int) -> tuple[int, int]:
        return (CS_ARCH_X86, CS_MODE_32)


class i8086Arch(PwndbgArchitecture):
    """
    16-bit mode x86
    """

    max_instruction_size = 16
    instruction_alignment = 1

    def __init__(self) -> None:
        super().__init__("i8086")

    @override
    def get_capstone_constants(self, address: int) -> tuple[int, int]:
        return (CS_ARCH_X86, CS_MODE_16)


class ArmArch(PwndbgArchitecture):
    max_instruction_size = 4
    instruction_alignment = 4

    def __init__(self) -> None:
        super().__init__("arm")

    @override
    def get_capstone_constants(self, address: int) -> tuple[int, int]:
        thumb_mode = pwndbg.aglib.disasm.disassembly.emulated_arm_mode_cache[address]
        if thumb_mode is None:
            thumb_mode = self.read_thumb_bit()
        mode = CS_MODE_THUMB if thumb_mode else CS_MODE_ARM

        return (CS_ARCH_ARM, mode)

    @override
    def read_thumb_bit(self) -> Literal[0, 1]:
        # When program initially starts, cpsr may not be readable
        if (cpsr := pwndbg.aglib.regs.read_reg("cpsr")) is not None:
            return (cpsr >> 5) & 1  # type: ignore[return-value]

        return 0


class ArmCortexArch(PwndbgArchitecture):
    """
    Cortex-M processors run the M-profile Arm architecture.
    This architecture is prevalent in bare-metal/embedded systems that lack operating systems.
    Only Thumb-2 instructions are supported, and the Thumb bit is always 1.
    """

    max_instruction_size = 4
    instruction_alignment = 4

    def __init__(self) -> None:
        super().__init__("armcm")

    @override
    def get_capstone_constants(self, address: int) -> tuple[int, int]:
        return (CS_ARCH_ARM, CS_MODE_MCLASS | CS_MODE_THUMB)

    @override
    def read_thumb_bit(self) -> Literal[0, 1]:
        """
        On Cortex-M processors, the Thumb bit is architecturally defined to be 1.

        This is the (xpsr >> 24) & 1, which is always 1.
        """
        return 1


class AArch64Arch(PwndbgArchitecture):
    max_instruction_size = 4
    instruction_alignment = 4
    constant_instruction_size = True

    def __init__(self) -> None:
        super().__init__("aarch64")

    @override
    def get_capstone_constants(self, address: int) -> tuple[int, int]:
        return (CS_ARCH_AARCH64, CS_MODE_ARM)

    @override
    def get_capstone_endianness(self) -> int:
        return CS_MODE_LITTLE_ENDIAN


class PowerPCArch(PwndbgArchitecture):
    max_instruction_size = 4
    instruction_alignment = 4

    def __init__(self) -> None:
        super().__init__("powerpc")

    @override
    def get_capstone_constants(self, address: int) -> tuple[int, int]:
        return (CS_ARCH_PPC, CS_MODE_64)


class SparcArch(PwndbgArchitecture):
    max_instruction_size = 4
    instruction_alignment = 4

    def __init__(self) -> None:
        super().__init__("sparc")

    @override
    def get_capstone_constants(self, address: int) -> tuple[int, int]:
        mode = CS_MODE_V9 if self.ptrsize == 8 else 0
        return (CS_ARCH_SPARC, mode)


class RISCV32Arch(PwndbgArchitecture):
    max_instruction_size = 22
    instruction_alignment = 4

    def __init__(self) -> None:
        super().__init__("rv32")

    @override
    def get_capstone_constants(self, address: int) -> tuple[int, int]:
        return (CS_ARCH_RISCV, CS_MODE_RISCV32 | CS_MODE_RISCVC)


class RISCV64Arch(PwndbgArchitecture):
    max_instruction_size = 22
    instruction_alignment = 4

    def __init__(self) -> None:
        super().__init__("rv64")

    @override
    def get_capstone_constants(self, address: int) -> tuple[int, int]:
        return (CS_ARCH_RISCV, CS_MODE_RISCV64 | CS_MODE_RISCVC)


class MipsArch(PwndbgArchitecture):
    max_instruction_size = 8
    instruction_alignment = 4

    def __init__(self) -> None:
        super().__init__("mips")

    @override
    def get_capstone_constants(self, address: int) -> tuple[int, int]:
        extra = 0
        for attribute in self.attributes:
            if attribute.cs_mode is not None:
                extra |= attribute.cs_mode

        if extra == 0:
            extra = CS_MODE_MIPS64 if self.ptrsize == 8 else CS_MODE_MIPS32

        return (CS_ARCH_MIPS, extra)


class Loongarch64Arch(PwndbgArchitecture):
    max_instruction_size = 4
    instruction_alignment = 4
    constant_instruction_size = True

    def __init__(self) -> None:
        super().__init__("loongarch64")

    @override
    def get_capstone_constants(self, address: int) -> tuple[int, int]:
        return (CS_ARCH_LOONGARCH, CS_MODE_LOONGARCH64)


class S390xArch(PwndbgArchitecture):
    max_instruction_size = 6
    instruction_alignment = 2

    def __init__(self) -> None:
        super().__init__("s390x")

    @override
    def get_capstone_constants(self, address: int) -> tuple[int, int]:
        return (CS_ARCH_SYSTEMZ, 0)


# Register the architecture classes
all_arches = [
    AMD64Arch(),
    i386Arch(),
    i8086Arch(),
    ArmArch(),
    ArmCortexArch(),
    AArch64Arch(),
    PowerPCArch(),
    SparcArch(),
    RISCV32Arch(),
    RISCV64Arch(),
    MipsArch(),
    Loongarch64Arch(),
    S390xArch(),
]

for arch in all_arches:
    register_arch(arch)


def get_thumb_mode_string() -> Literal["arm", "thumb"] | None:
    thumb_bit = pwndbg.aglib.arch.read_thumb_bit()
    return None if thumb_bit is None else "thumb" if thumb_bit == 1 else "arm"


def update() -> None:
    a = pwndbg.dbg.selected_inferior().arch()

    pwnlib.context.context.arch = PWNLIB_ARCH_MAPPINGS.get(a.name, "none")
    pwnlib.context.context.bits = a.ptrsize * 8
    pwnlib.context.context.endian = a.endian
    pwnlib.context.context.os = PWNLIB_PLATFORM_MAPPINGS.get(a.platform, "linux")

    # Will be None the first time around.
    if pwndbg.aglib.arch is None or a.name != pwndbg.aglib.arch.name:
        pwndbg_arch = get_pwndbg_architecture(a.name)
        if pwndbg_arch is None:
            raise pwndbg.dbg_mod.Error(
                f"Unsupported architecture: {a.name}. "
                f"It may be that Pwndbg is not correctly categorizing the architecture. "
                f"Please file a bug report. "
            )
        pwndbg.aglib.set_arch(pwndbg_arch)

    pwndbg.aglib.arch.update(a)



from __future__ import annotations

from typing import Any


class ABI:
    """
    Encapsulates information about a calling convention.
    """

    #: Tuple of registers which should be filled with arguments before
    #: spilling onto the stack.
    register_arguments: tuple[str, ...] = ()

    #: Minimum alignment of the stack.
    #: The value used is min(context.bytes, stack_alignment)
    #: This is necessary as Windows x64 frames must be 32-byte aligned.
    #: "Alignment" is considered with respect to the last argument on the stack.
    arg_alignment = 1

    #: Minimum number of stack slots used by a function call
    #: This is necessary as Windows x64 requires using 4 slots on the stack
    stack_minimum = 0

    #: Indicates that this ABI returns to the next address on the slot
    returns = True

    def __init__(self, regs: tuple[str, ...], align: int, minimum: int) -> None:
        self.register_arguments = regs
        self.arg_alignment = align
        self.stack_minimum = minimum


class SyscallABI(ABI):
    """
    The syscall ABI treats the syscall number as the zeroth argument,
    which must be loaded into the specified register.
    """

    def __init__(self, register_arguments: tuple[str, ...], *a: Any, **kw: Any) -> None:
        self.syscall_register = register_arguments[0]
        super().__init__(register_arguments[1:], *a, **kw)


class SigreturnABI(SyscallABI):
    """
    The sigreturn ABI is similar to the syscall ABI, except that
    both PC and SP are loaded from the stack.  Because of this, there
    is no 'return' slot necessary on the stack.
    """

    returns = False


linux_i386 = ABI((), 4, 0)
linux_amd64 = ABI(("rdi", "rsi", "rdx", "rcx", "r8", "r9"), 8, 0)
linux_arm = ABI(("r0", "r1", "r2", "r3"), 8, 0)
linux_aarch64 = ABI(("x0", "x1", "x2", "x3", "x4", "x5", "x6", "x7"), 16, 0)
linux_mips = ABI(("$a0", "$a1", "$a2", "$a3"), 4, 0)
linux_mips64 = ABI(("$a0", "$a1", "$a2", "$a3", "$a4", "$a5", "$a6", "$a7"), 8, 0)
linux_ppc = ABI(("r3", "r4", "r5", "r6", "r7", "r8", "r9", "r10"), 4, 0)
linux_ppc64 = ABI(("r3", "r4", "r5", "r6", "r7", "r8", "r9", "r10"), 8, 0)
linux_riscv32 = ABI(("a0", "a1", "a2", "a3", "a4", "a5", "a6", "a7"), 4, 0)
linux_riscv64 = ABI(("a0", "a1", "a2", "a3", "a4", "a5", "a6", "a7"), 8, 0)

linux_i386_syscall = SyscallABI(("eax", "ebx", "ecx", "edx", "esi", "edi", "ebp"), 4, 0)
linux_amd64_syscall = SyscallABI(("rax", "rdi", "rsi", "rdx", "r10", "r8", "r9"), 8, 0)
linux_arm_syscall = SyscallABI(("r7", "r0", "r1", "r2", "r3", "r4", "r5", "r6"), 4, 0)
linux_aarch64_syscall = SyscallABI(("x8", "x0", "x1", "x2", "x3", "x4", "x5"), 16, 0)
linux_mips_syscall = SyscallABI(("$v0", "$a0", "$a1", "$a2", "$a3"), 4, 0)
linux_mips64_syscall = SyscallABI(("$v0", "$a0", "$a1", "$a2", "$a3", "$a4", "$a5"), 4, 0)
linux_ppc_syscall = SyscallABI(("r0", "r3", "r4", "r5", "r6", "r7", "r8", "r9"), 4, 0)
linux_ppc64_syscall = SyscallABI(("r0", "r3", "r4", "r5", "r6", "r7", "r8"), 8, 0)
linux_riscv32_syscall = SyscallABI(("a7", "a0", "a1", "a2", "a3", "a4", "a5", "a6"), 4, 0)
linux_riscv64_syscall = SyscallABI(("a7", "a0", "a1", "a2", "a3", "a4", "a5", "a6"), 8, 0)

linux_i386_sigreturn = SigreturnABI(("eax",), 4, 0)
linux_amd64_sigreturn = SigreturnABI(("rax",), 4, 0)
linux_arm_sigreturn = SigreturnABI(("r7",), 4, 0)

# Fake ABIs used by SROP
linux_i386_srop = ABI(("eax",), 4, 0)
linux_amd64_srop = ABI(("rax",), 4, 0)
linux_arm_srop = ABI(("r7",), 4, 0)

DEFAULT_ABIS: dict[tuple[int, str, str], ABI] = {
    (32, "i386", "linux"): linux_i386,
    (64, "x86-64", "linux"): linux_amd64,
    (64, "aarch64", "linux"): linux_aarch64,
    (32, "arm", "linux"): linux_arm,
    (32, "thumb", "linux"): linux_arm,
    (32, "mips", "linux"): linux_mips,
    (64, "mips", "linux"): linux_mips64,
    (32, "powerpc", "linux"): linux_ppc,
    (64, "powerpc", "linux"): linux_ppc64,
    (32, "rv32", "linux"): linux_riscv32,
    (64, "rv64", "linux"): linux_riscv64,
}

SYSCALL_ABIS: dict[tuple[int, str, str], SyscallABI] = {
    (32, "i386", "linux"): linux_i386_syscall,
    (64, "x86-64", "linux"): linux_amd64_syscall,
    (64, "aarch64", "linux"): linux_aarch64_syscall,
    (32, "arm", "linux"): linux_arm_syscall,
    (32, "thumb", "linux"): linux_arm_syscall,
    (32, "mips", "linux"): linux_mips_syscall,
    (64, "mips", "linux"): linux_mips64_syscall,
    (32, "powerpc", "linux"): linux_ppc_syscall,
    (64, "powerpc", "linux"): linux_ppc64_syscall,
    (32, "rv32", "linux"): linux_riscv32_syscall,
    (64, "rv64", "linux"): linux_riscv64_syscall,
}

SIGRETURN_ABIS: dict[tuple[int, str, str], SigreturnABI] = {
    (32, "i386", "linux"): linux_i386_sigreturn,
    (64, "x86-64", "linux"): linux_amd64_sigreturn,
    (32, "arm", "linux"): linux_arm_sigreturn,
    (32, "thumb", "linux"): linux_arm_sigreturn,
}
